from __future__ import annotations

from collections.abc import Callable, Container
import logging

from prettyqt import core, eventfilters
from prettyqt.qt import QtCore


logger = logging.getLogger(__name__)


class EventCatcher(eventfilters.BaseEventFilter):
    caught = core.Signal(QtCore.QEvent)

    def __init__(
        self,
        include: QtCore.QEvent.Type
        | core.Event.TypeStr
        | Container[QtCore.QEvent.Type | core.Event.TypeStr]
        | None = None,
        exclude: QtCore.QEvent.Type
        | core.Event.TypeStr
        | Container[QtCore.QEvent.Type | core.Event.TypeStr]
        | None = None,
        do_filter: bool | Callable[[QtCore.QEvent], bool] = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        match include:
            case str():
                self.include = [core.event.TYPE[include]]
            case Container():
                self.include = [
                    core.event.TYPE[i] if isinstance(i, str) else i for i in include
                ]
            case QtCore.QEvent.Type():
                self.include = [include]
            case None:
                self.include = None
            case _:
                raise ValueError(include)
        match exclude:
            case str():
                self.exclude = [core.event.TYPE[exclude]]

            case Container():
                self.exclude = [
                    core.event.TYPE[i] if isinstance(i, str) else i for i in exclude
                ]
            case QtCore.QEvent.Type():
                self.exclude = [exclude]
            case None:
                self.exclude = None
            case _:
                raise ValueError(exclude)
        self.do_filter = do_filter

    def eventFilter(self, source: QtCore.QObject, event: QtCore.QEvent) -> bool:
        if not self.include or event.type() in self.include:
            if not self.exclude or event.type() not in self.exclude:
                self.caught.emit(event)
                logger.debug(f"{source!r}: {event.type()!r}")
                if callable(self.do_filter):
                    return self.do_filter(event)
                return self.do_filter
        return False


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    window = widgets.MainWindow()
    eventcatcher = EventCatcher()



"""designer module.

contains QtDesigner-based classes
"""
from prettyqt import qt

if qt.API == "pyqt6":
    from .pydesignercustomwidgetcollectionplugin import (
        PyDesignerCustomWidgetCollectionPlugin,
    )
    from .pydesignercustomwidgetplugin import PyDesignerCustomWidgetPlugin
else:
    from .pydesignercustomwidgetcollection import PyDesignerCustomWidgetCollection

from .abstractextensionfactory import AbstractExtensionFactory
from .pydesignertaskmenuextension import PyDesignerTaskMenuExtension

__all__ = [
    "PyDesignerCustomWidgetPlugin",
    "PyDesignerCustomWidgetCollectionPlugin",
    "PyDesignerCustomWidgetCollection",
    "AbstractExtensionFactory",
    "PyDesignerTaskMenuExtension",
]



from __future__ import annotations

from re import Pattern

from prettyqt import custom_validators, custom_widgets, widgets
from prettyqt.qt import QtWidgets


class RegexLineEdit(custom_widgets.SingleLineTextEdit):
    def __init__(
        self,
        parent: QtWidgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        self.set_syntaxhighlighter("regex")
        self.val = custom_validators.RegexPatternValidator()
        self.set_margin(0)
        self.set_validator(self.val)

    @property
    def pattern(self) -> str:
        return self.text()

    @pattern.setter
    def pattern(self, value: str):
        self.set_text(value)

    def set_value(self, value: str | Pattern | None):
        match value:
            case None:
                self.pattern = ""
            case str():
                self.pattern = value
            case _:
                self.pattern = value.pattern


if __name__ == "__main__":
    app = widgets.app()
    widget = RegexLineEdit()
    widget.show()
    widget.value_changed.connect(print)
    app.main_loop()



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.qt import QtCore, QtSvg


class SvgGenerator(gui.PaintDeviceMixin, QtSvg.QSvgGenerator):
    def get_viewbox(self) -> core.Rect:
        return core.Rect(self.viewBox())

    def get_viewboxf(self) -> core.RectF:
        return core.RectF(self.viewBoxF())

    def get_size(self) -> core.Size:
        return core.Size(self.size())

    def set_size(self, size: QtCore.QSize | QtCore.QSizeF | tuple[int, int]):
        match size:
            case tuple():
                new_size = QtCore.QSize(*size)
            case QtCore.QSizeF():
                new_size = size.toSize()
            case QtCore.QSize():
                new_size = size
        self.setSize(new_size)



from __future__ import annotations

from collections.abc import Iterator
import contextlib
import os
import pathlib
from typing import Literal

import qstylizer.parser
import qstylizer.style

from prettyqt import constants, core, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, bidict, datatypes, get_repr


MARKDOWN_FEATURES = bidict(
    no_html=QtGui.QTextDocument.MarkdownFeature.MarkdownNoHTML,
    commonmark=QtGui.QTextDocument.MarkdownFeature.MarkdownDialectCommonMark,
    github=QtGui.QTextDocument.MarkdownFeature.MarkdownDialectGitHub,
)

MarkdownFeatureStr = Literal["no_html", "commonmark", "github"]

RESOURCE_TYPES = bidict(
    unknown=QtGui.QTextDocument.ResourceType.UnknownResource,
    html=QtGui.QTextDocument.ResourceType.HtmlResource,
    image=QtGui.QTextDocument.ResourceType.ImageResource,
    stylesheet=QtGui.QTextDocument.ResourceType.StyleSheetResource,
    markdown=QtGui.QTextDocument.ResourceType.MarkdownResource,
    user=QtGui.QTextDocument.ResourceType.UserResource,
)

ResourceTypeStr = Literal["unknown", "html", "image", "stylesheet", "markdown", "user"]

STACKS = bidict(
    undo=QtGui.QTextDocument.Stacks.UndoStack,
    redo=QtGui.QTextDocument.Stacks.RedoStack,
    undo_and_redo=QtGui.QTextDocument.Stacks.UndoAndRedoStacks,
)

StackStr = Literal["undo", "redo", "undo_and_redo"]

FIND_FLAGS = bidict(
    backward=QtGui.QTextDocument.FindFlag.FindBackward,
    case_sensitive=QtGui.QTextDocument.FindFlag.FindCaseSensitively,
    whole_words=QtGui.QTextDocument.FindFlag.FindWholeWords,
)

FindFlagStr = Literal["backward", "case_sensitive", "whole_words"]

META_INFORMATION = bidict(
    document_title=QtGui.QTextDocument.MetaInformation.DocumentTitle,
    document_url=QtGui.QTextDocument.MetaInformation.DocumentUrl,
    css_media=QtGui.QTextDocument.MetaInformation.CssMedia,
)

MetaInformationStr = Literal["document_title", "document_url", "css_media"]


class TextDocumentMixin(core.ObjectMixin):
    def __getitem__(self, index: int) -> gui.TextBlock:
        return gui.TextBlock(self.findBlockByNumber(index))

    def __len__(self) -> int:
        return self.blockCount()

    def __iter__(self) -> Iterator[gui.TextBlock]:
        return iter(
            gui.TextBlock(self.findBlockByNumber(i)) for i in range(self.blockCount())
        )

    def __repr__(self):
        return get_repr(self, self.toPlainText())

    def get_first_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.firstBlock())

    def get_last_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.lastBlock())

    def find_block_by_number(self, number: int) -> gui.TextBlock:
        block = self.findBlockByNumber(number)
        if not block.isValid():
            raise ValueError(
                f"{number} not a valid block index. Block count: {self.blockCount()}"
            )
        return gui.TextBlock(block)

    def find_block_by_line_number(self, line_number: int) -> gui.TextBlock:
        block = self.findBlockByLineNumber(line_number)
        if not block.isValid():
            raise ValueError(
                f"{line_number} not a valid line index. Line count: {self.lineCount()}"
            )
        return gui.TextBlock(block)

    def set_text(self, text: str):
        self.setPlainText(text)

    def serialize_fields(self):
        return dict(
            base_url=self.get_base_url(),
            default_font=self.get_default_font(),
            default_stylesheet=self.defaultStyleSheet(),
            default_text_option=self.get_default_text_option(),
            document_margin=self.documentMargin(),
            maximum_block_count=self.maximumBlockCount(),
            is_modified=self.isModified(),
            page_size=self.pageSize(),
            text_width=self.textWidth(),
            indent_width=self.indentWidth(),
            undo_redo_enabled=self.isUndoRedoEnabled(),
            use_design_metrics=self.useDesignMetrics(),
        )

    def get_base_url(self) -> core.Url:
        return core.Url(self.baseUrl())

    def get_default_font(self) -> gui.Font:
        return gui.Font(self.defaultFont())

    def set_default_text_option(self, opt: QtGui.QTextOption):
        self.setDefaultTextOption(gui.TextOption(opt))

    def get_default_text_option(self) -> gui.TextOption:
        return gui.TextOption(self.defaultTextOption())

    def set_flags(self, **flags):
        current = self.flags()
        for k, v in flags.items():
            if v:
                current = current | gui.textoption.FLAG[k]
            else:
                current = current & ~gui.textoption.FLAG[k]
        self.setFlags(current)
        # if show:
        #     self.setFlags(self.flags() | QtGui.QTextOption.ShowTabsAndSpaces)
        # else:
        #     self.setFlags(self.flags() & ~QtGui.QTextOption.ShowTabsAndSpaces)

    def clear_stacks(self, stack: StackStr):
        """Clear undo / redo stack.

        Args:
            stack: stack to clear

        Raises:
            InvalidParamError: stack type does not exist
        """
        if stack not in STACKS:
            raise InvalidParamError(stack, STACKS)
        self.clearUndoRedoStacks(STACKS[stack])

    def set_default_cursor_move_style(self, style: constants.CursorMoveStyleStr):
        """Set the cursor move style.

        Args:
            style: cursor move style

        Raises:
            InvalidParamError: cursor move style does not exist
        """
        if style not in constants.CURSOR_MOVE_STYLE:
            raise InvalidParamError(style, constants.CURSOR_MOVE_STYLE)
        self.setDefaultCursorMoveStyle(constants.CURSOR_MOVE_STYLE[style])

    def get_default_cursor_move_style(self) -> constants.CursorMoveStyleStr:
        """Return current cursor move style.

        Returns:
            cursor move style
        """
        return constants.CURSOR_MOVE_STYLE.inverse[self.defaultCursorMoveStyle()]

    def set_meta_information(self, info: MetaInformationStr, value: str):
        """Set meta information.

        Args:
            info: meta information type
            value: value to set

        Raises:
            InvalidParamError: meta information type does not exist
        """
        if info not in META_INFORMATION:
            raise InvalidParamError(info, META_INFORMATION)
        self.setMetaInformation(META_INFORMATION[info], value)

    def get_meta_information(self, info: MetaInformationStr) -> str:
        """Return specififed meta information.

        Args:
            info: meta information type

        Returns:
            meta information
        """
        if info not in META_INFORMATION:
            raise InvalidParamError(info, META_INFORMATION)
        return self.metaInformation(META_INFORMATION[info])

    def add_resource(
        self, resource_type: ResourceTypeStr, name: datatypes.PathType, resource
    ):
        if resource_type not in RESOURCE_TYPES:
            raise InvalidParamError(resource_type, RESOURCE_TYPES)
        url = core.Url(name)
        self.addResource(RESOURCE_TYPES[resource_type], url, resource)

    @contextlib.contextmanager
    def edit_default_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_default_stylesheet()
        yield ss
        self.set_default_stylesheet(ss)

    def set_default_stylesheet(
        self, ss: None | str | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        if isinstance(ss, os.PathLike):
            ss = pathlib.Path(ss).read_text()
        elif ss is None:
            ss = ""
        self.setDefaultStyleSheet(str(ss))

    def get_default_stylesheet(self) -> qstylizer.style.StyleSheet:
        return qstylizer.parser.parse(self.defaultStyleSheet())

    def find_line_position(self, line_no: int) -> int:
        """Return index of the TextBlock corresponding to given line number."""
        lines = self.blockCount()
        assert 1 <= line_no <= lines
        return self.findBlockByLineNumber(line_no - 1).position()

    def write_to_file(
        self,
        path: datatypes.PathType,
        fmt: gui.textdocumentwriter.FormatStr | bytes | QtCore.QByteArray = "plaintext",
    ) -> bool:
        writer = gui.TextDocumentWriter()
        writer.set_format(fmt)
        writer.set_file_name(path)
        return writer.write(self)

    def get_bytes(
        self, fmt: Literal["plaintext", "HTML", "markdown", "ODF"] = "ODF"
    ) -> bytes:
        return gui.TextDocumentWriter.serialize_document(self, fmt)

    def show_whitespace_and_tabs(self, show: bool):
        """Set show white spaces flag."""
        options = self.get_default_text_option()
        flag = QtGui.QTextOption.Flag.ShowTabsAndSpaces
        if show:
            options.setFlags(options.flags() | flag)  # type: ignore
        else:
            options.setFlags(options.flags() & ~flag)  # type: ignore
        self.setDefaultTextOption(options)


class TextDocument(TextDocumentMixin, QtGui.QTextDocument):
    pass


if __name__ == "__main__":
    doc = TextDocument("This is a test\nHello")
    doc.set_default_text_option(QtGui.QTextOption())
    a = doc.get_bytes()
    print(a)



"""Custom_models module.

Contains custom models
"""

from .selectionmixin import SelectionMixin
from .listmixin import ListMixin
from .treemodel import TreeModel
from .columnitemmodel import ColumnItemModel, ColumnTableModel, ColumnItem
from .regexmatchesmodel import RegexMatchesModel
from .importlibdistributionmodel import ImportlibDistributionModel
from .modelmixin import ModelMixin
from .nesteditem import NestedItem
from .nestedmodel import NestedModel
from .jsonmodel import JsonModel
from .subsequencesortfilterproxymodel import SubsequenceSortFilterProxyModel
from .fuzzyfiltermodel import FuzzyFilterProxyModel

__all__ = [
    "SelectionMixin",
    "TreeModel",
    "ImportlibDistributionModel",
    "RegexMatchesModel",
    "ColumnItemModel",
    "ColumnTableModel",
    "ListMixin",
    "ModelMixin",
    "ColumnItem",
    "NestedModel",
    "NestedItem",
    "JsonModel",
    "SubsequenceSortFilterProxyModel",
    "FuzzyFilterProxyModel",
]



"""Module containing custom delegate classes."""

from .buttondelegate import ButtonDelegate
from .radiodelegate import RadioDelegate
from .progressbardelegate import ProgressBarDelegate
from .icondelegate import IconDelegate
from .stardelegate import StarDelegate
from .renderlinkdelegate import RenderLinkDelegate
from .nofocusdelegate import NoFocusDelegate
from .htmlitemdelegate import HtmlItemDelegate
from .widgetdelegate import WidgetDelegate

__all__ = [
    "ButtonDelegate",
    "RadioDelegate",
    "ProgressBarDelegate",
    "IconDelegate",
    "StarDelegate",
    "RenderLinkDelegate",
    "NoFocusDelegate",
    "HtmlItemDelegate",
    "WidgetDelegate",
]



from __future__ import annotations

import contextlib
import functools
import os
import html
import pathlib
import sys
from typing import TYPE_CHECKING, Any, Literal

import qstylizer.parser
import qstylizer.style

from prettyqt import constants, core, gui, iconprovider, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import InvalidParamError, colors, datatypes, prettyprinter


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator


LayoutStr = Literal["horizontal", "vertical", "grid", "form", "stacked", "flow"]

QWIDGETSIZE_MAX = 16777215  # QtWidgets.QWIDGETSIZE_MAX


class WidgetMixin(core.ObjectMixin):
    box: QtWidgets.QLayout

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "contextMenuPolicy": constants.CONTEXT_POLICY,
            "focusPolicy": constants.FOCUS_POLICY,
            "layoutDirection": constants.LAYOUT_DIRECTION,
            "windowModality": constants.WINDOW_MODALITY,
        }
        return maps

    def add_shortcut(
        self,
        keysequence: datatypes.KeyCombinationType,
        callback: Callable | None = None,
        context: constants.ShortcutContextStr = "window",
    ) -> gui.Shortcut:
        """Add shortcut to widget.

        Adds as shortcut for given callback and context to this widget.

        Args:
            keysequence: Key sequence
            callback: Callback for the shortcut
            context: context for this shortcut

        Returns:
            shortcut object
        """
        if not isinstance(keysequence, QtGui.QKeySequence):
            keysequence = gui.KeySequence(keysequence)
        context = constants.SHORTCUT_CONTEXT[context]
        return gui.Shortcut(keysequence, self, callback, context=context)

    def get_win_id(self) -> int:
        return int(self.winId())

    def resize(self, *size) -> None:
        if isinstance(size[0], tuple):
            super().resize(*size[0])
        else:
            super().resize(*size)

    def insertAction(
        self, position_or_action: int | QtGui.QAction, action: QtGui.QAction
    ):
        """Extend insertAction to also allow int index."""
        if isinstance(position_or_action, int):
            position_or_action = self.actions()[position_or_action]
        super().insertAction(position_or_action, action)

    def add_action(
        self,
        text: str | gui.Action,
        triggered: Callable = None,
        icon: Any | None = None,
        checkable: bool = False,
        checked: bool = False,
        shortcut: str | None = None,
        status_tip: str | None = None,
        tool_tip: str | None = None,
        parent=None,
    ) -> gui.Action:
        """Add an action to the menu.

        Args:
            text: Label for the action
            triggered: gets called when action is triggered
            icon: icon for action
            checkable: as checkbox action
            checked: if checkable, turn on by default
            shortcut: Shortcut for action
            status_tip: Status tip to be shown in status bar
            tool_tip: Widget ToolTip
            parent: parent

        Returns:
            Action added to menu
        """
        if isinstance(text, str):
            action = gui.Action(parent or self, text=text)
            if triggered:
                action.triggered.connect(triggered)
            action.set_icon(icon)
            action.set_shortcut(shortcut)
            if checkable:
                action.setCheckable(True)
                action.setChecked(checked)
            if status_tip is not None:
                action.setStatusTip(status_tip)
            if tool_tip is not None:
                action.setToolTip(tool_tip)
        else:
            action = text
            action.setParent(self)
        self.addAction(action)
        return action

    def add_actions(self, actions: list[QtGui.QAction]):
        for i in actions:
            i.setParent(self)
        self.addActions(actions)

    def toggle_fullscreen(self):
        """Toggle between fullscreen and regular size."""
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def toggle_maximized(self):
        """Toggle between maximized and regular size."""
        if self.isMaximized():
            self.showNormal()
        else:
            self.showMaximized()

    def set_graphics_effect(
        self,
        effect: QtWidgets.QGraphicsEffect
        | Literal["drop_shadow", "blur", "opacity", "colorize"],
        radius: int = 10,
        opacity: int = 0.7,
        strength: int = 0.5,
        color: datatypes.ColorType = "blue",
    ) -> QtWidgets.QGraphicsEffect:
        match effect:
            case "drop_shadow":
                effect = widgets.GraphicsDropShadowEffect(self)
                effect.setBlurRadius(radius)
                effect.setColor(colors.get_color(color))
            case "blur":
                effect = widgets.GraphicsBlurEffect(self)
                effect.setBlurRadius(radius)
            case "opacity":
                effect = widgets.GraphicsOpacityEffect(self)
                effect.setOpacity(opacity)
            case "colorize":
                effect = widgets.GraphicsColorizeEffect(self)
                effect.setColor(colors.get_color(color))
                effect.setStrength(strength)
        self.setGraphicsEffect(effect)
        return effect

    def play_animation(self, animation_type: str, **kwargs) -> QtCore.QPropertyAnimation:
        from prettyqt import custom_animations

        match animation_type:
            case "fade_in":
                anim = custom_animations.FadeInAnimation(**kwargs, parent=self)
                anim.apply_to(self)
            case "bounce":
                anim = custom_animations.BounceAnimation(**kwargs, parent=self)
                anim.apply_to(self)
            case "slide":
                anim = custom_animations.SlideAnimation(**kwargs, parent=self)
                anim.apply_to(self)
            case "property":
                anim = core.PropertyAnimation(parent=self)
                anim.setTargetObject(self)
                anim.set_property_name(kwargs.pop("name"))
                anim.setStartValue(kwargs.pop("start_value"))
                anim.setEndValue(kwargs.pop("end_value"))
                anim.setDuration(kwargs.pop("duration", 1000))
                anim.set_easing(kwargs.pop("easing", "linear"))
        anim.start()
        return anim

    def map_to_global(self, pos_or_rect):
        match pos_or_rect:
            case QtCore.QRect() | QtCore.QRectF():
                top_left = self.mapToGlobal(pos_or_rect.topLeft())
                bottom_right = self.mapToGlobal(pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case QtCore.QPoint() | QtCore.QPointF():
                return super().mapToGlobal(pos_or_rect)
            case int(), int():
                return self.mapToGlobal(QtCore.QPoint(*pos_or_rect))
            case float(), float():
                return self.mapToGlobal(QtCore.QPointF(*pos_or_rect))
            case int(), int(), int(), int():
                return self.map_to_global(QtCore.QRect(*pos_or_rect))
            case float(), float(), float(), float():
                return self.map_to_global(QtCore.QRectF(*pos_or_rect))
            case _:
                raise ValueError(pos_or_rect)

    def raise_to_top(self):
        if sys.platform.startswith("win"):
            import win32con
            from win32gui import SetWindowPos

            # set to always-on-top and disable it again. that way windows stays in front
            flag = win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_SHOWWINDOW
            win_id = self.winId()
            SetWindowPos(win_id, win32con.HWND_TOPMOST, 0, 0, 0, 0, flag)
            SetWindowPos(win_id, win32con.HWND_NOTOPMOST, 0, 0, 0, 0, flag)
        # state = (self.windowState() & ~Qt.WindowMinimized) | Qt.WindowActive
        # self.setWindowState(state)
        self.raise_()
        self.show()
        self.activateWindow()

    def set_icon(self, icon: datatypes.IconType) -> None:
        """Set the window icon.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon, color=colors.WINDOW_ICON_COLOR)
        super().setWindowIcon(icon)

    setWindowIcon = set_icon

    def get_icon(self) -> gui.Icon | None:
        icon = super().windowIcon()
        return None if icon.isNull() else gui.Icon(icon)

    def show_tooltip(self, duration: int | None = None):
        if duration is None:
            duration = -1  # automatic
        widgets.ToolTip.showText(
            self.map_to_global((0, 0)), self.toolTip(), msecShowTime=duration
        )

    @functools.singledispatchmethod
    def set_min_size(self, size: QtCore.QSize | tuple[int | None, int | None]) -> None:
        if isinstance(size, tuple):
            x = 0 if size[0] is None else size[0]
            y = 0 if size[1] is None else size[1]
            super().setMinimumSize(x, y)
        else:
            super().setMinimumSize(size)

    setMinimumSize = set_min_size

    @set_min_size.register
    def _(self, x: int, y: int | None):
        self.set_min_size((x, y))

    @set_min_size.register  # these can be merged when min py version is 3.11
    def _(self, x: None, y: int | None):
        self.set_min_size((x, y))

    @functools.singledispatchmethod
    def set_max_size(self, size: QtCore.QSize | tuple[int | None, int | None]) -> None:
        if isinstance(size, tuple):
            x = QWIDGETSIZE_MAX if size[0] is None else size[0]
            y = QWIDGETSIZE_MAX if size[1] is None else size[1]
            super().setMaximumSize(x, y)
        else:
            super().setMaximumSize(size)

    setMaximumSize = set_min_size

    @set_max_size.register
    def _(self, x: int, y: int | None):
        self.set_max_size((x, y))

    @set_max_size.register  # these can be merged when min py version is 3.11
    def _(self, x: None, y: int | None):
        self.set_max_size((x, y))

    def set_min_width(self, width: int | None) -> None:
        if width is None:
            width = 0
        super().setMinimumWidth(width)

    setMinimumWidth = set_min_width

    def set_max_width(self, width: int | None) -> None:
        if width is None:
            width = QWIDGETSIZE_MAX
        super().setMaximumWidth(width)

    setMaximumWidth = set_max_width

    def set_min_height(self, height: int | None) -> None:
        if height is None:
            height = 0
        super().setMinimumHeight(height)

    setMinimumHeight = set_min_height

    def set_max_height(self, height: int | None) -> None:
        if height is None:
            height = QWIDGETSIZE_MAX
        super().setMaximumHeight(height)

    setMaximumHeight = set_max_height

    def set_enabled(self, enabled: bool = True) -> None:
        self.setEnabled(enabled)

    def set_disabled(self) -> None:
        self.setEnabled(False)

    def setWindowTitle(self, title: str):
        if not self.objectName() and widgets.app().is_debug():
            self.setObjectName(title)
        super().setWindowTitle(title)

    def set_title(self, title: str) -> None:
        self.setWindowTitle(title)

    def get_title(self) -> str:
        return self.windowTitle()

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        size: datatypes.SizeType | None = None,
        rich_text: bool = False,
    ):
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                if isinstance(size, QtCore.QSize):
                    size = (size.width(), size.height())
                tooltip = f'<img src={path!r} width="{size[0]}" height="{size[1]}">'
        tooltip = tooltip.replace("\n", "<br/>")
        if rich_text:
            tooltip = f"<html>{html.escape(tooltip)}</html>"
        super().setToolTip(tooltip)

    def set_font(
        self,
        font_name: str | None = None,
        font_size: int | None = None,
        weight: int | None = None,
        italic: bool = False,
    ) -> gui.Font:
        if font_size is None:
            font_size = -1
        if weight is None:
            weight = -1
        if font_name is None:
            font_name = self.font().family()
        font = gui.Font(font_name, font_size, weight, italic)
        super().setFont(font)
        return font

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_foreground_role(self) -> gui.palette.RoleStr:
        return gui.palette.ROLE.inverse[self.foregroundRole()]

    def set_foreground_role(self, role: gui.palette.RoleStr):
        if role not in gui.palette.ROLE:
            raise InvalidParamError(role, gui.palette.ROLE)
        self.setForegroundRole(gui.palette.ROLE[role])

    def get_background_role(self) -> gui.palette.RoleStr:
        return gui.palette.ROLE.inverse[self.backgroundRole()]

    def set_background_role(self, role: gui.palette.RoleStr):
        if role not in gui.palette.ROLE:
            raise InvalidParamError(role, gui.palette.ROLE)
        self.setBackgroundRole(gui.palette.ROLE[role])

    def set_window_flags(self, *flags: constants.WindowTypeStr, append: bool = False):
        for flag in flags:
            if flag not in constants.WINDOW_TYPE:
                raise InvalidParamError(flag, constants.WINDOW_TYPE)
        result = constants.WINDOW_TYPE.merge_flags(flags)
        if append:
            result = result | self.windowFlags()
        self.setWindowFlags(result)

    def set_flags(
        self,
        minimize: bool | None = None,
        maximize: bool | None = None,
        close: bool | None = None,
        stay_on_top: bool | None = None,
        frameless: bool | None = None,
        window: bool | None = None,
        dialog: bool | None = None,
        tooltip: bool | None = None,
        tool: bool | None = None,
        customize: bool | None = None,
        window_title: bool | None = None,
    ) -> None:
        flags = {
            minimize: QtCore.Qt.WindowType.WindowMinimizeButtonHint,
            maximize: QtCore.Qt.WindowType.WindowMaximizeButtonHint,
            close: QtCore.Qt.WindowType.WindowCloseButtonHint,
            stay_on_top: QtCore.Qt.WindowType.WindowStaysOnTopHint,
            frameless: QtCore.Qt.WindowType.FramelessWindowHint,
            window: QtCore.Qt.WindowType.Window,
            dialog: QtCore.Qt.WindowType.Dialog,
            tooltip: QtCore.Qt.WindowType.ToolTip,
            tool: QtCore.Qt.WindowType.Tool,
            customize: QtCore.Qt.WindowType.CustomizeWindowHint,
            window_title: QtCore.Qt.WindowType.WindowTitleHint,
        }
        for k, v in flags.items():
            if k is not None:
                self.setWindowFlag(v, k)

    def set_attribute(
        self, attribute: constants.WidgetAttributeStr, state: bool = True
    ) -> None:
        if attribute not in constants.WIDGET_ATTRIBUTE:
            raise InvalidParamError(attribute, constants.WIDGET_ATTRIBUTE)
        self.setAttribute(constants.WIDGET_ATTRIBUTE[attribute], state)

    def set_attributes(self, **kwargs: bool) -> None:
        for attr, state in kwargs.items():
            if attr not in constants.WIDGET_ATTRIBUTE:
                raise InvalidParamError(attr, constants.WIDGET_ATTRIBUTE)
            self.setAttribute(constants.WIDGET_ATTRIBUTE[attr], state)  # type: ignore

    def set_modality(self, modality: constants.WindowModalityStr) -> None:
        """Set modality for the dialog.

        Args:
            modality: modality for the main window

        Raises:
            InvalidParamError: modality type does not exist
        """
        if modality not in constants.WINDOW_MODALITY:
            raise InvalidParamError(modality, constants.WINDOW_MODALITY)
        self.setWindowModality(constants.WINDOW_MODALITY[modality])

    def get_modality(self) -> constants.WindowModalityStr:
        """Get the current modality modes as a string.

        Returns:
            modality mode
        """
        return constants.WINDOW_MODALITY.inverse[self.windowModality()]

    def set_size_policy(
        self,
        horizontal: widgets.sizepolicy.SizePolicyStr | None = None,
        vertical: widgets.sizepolicy.SizePolicyStr | None = None,
    ) -> None:
        """Set the sizes policy.

        Args:
            horizontal: horizontal size policy
            vertical: vertical size policy
        """
        sp = self.get_size_policy()
        if horizontal is not None:
            sp.set_horizontal_policy(horizontal)
        if vertical is not None:
            sp.set_vertical_policy(vertical)
        self.setSizePolicy(sp)

    def get_size_policy(self) -> widgets.SizePolicy:
        qpol = self.sizePolicy()
        if isinstance(qpol, widgets.SizePolicy):
            return qpol
        return widgets.SizePolicy.clone(qpol)

    def get_palette(self) -> gui.Palette:
        return gui.Palette(self.palette())

    def set_background_color(self, color: datatypes.ColorType) -> None:
        col_str = "" if color is None else colors.get_color(color).name()
        with self.edit_stylesheet() as ss:
            ss.backgroundColor.setValue(col_str)

    @contextlib.contextmanager
    def grab_mouse_events(
        self, cursor_shape: constants.CursorShapeStr | None = None
    ) -> Iterator[None]:
        if cursor_shape is not None:
            self.grabMouse(constants.CURSOR_SHAPE[cursor_shape])
        else:
            self.grabMouse()
        yield None
        self.releaseMouse()

    @contextlib.contextmanager
    def grab_keyboard_events(self) -> Iterator[None]:
        self.grabKeyboard()
        yield None
        self.releaseKeyboard()

    @contextlib.contextmanager
    def updates_off(self) -> Iterator[None]:
        updates = self.updatesEnabled()
        self.setUpdatesEnabled(False)
        yield None
        self.setUpdatesEnabled(updates)

    @contextlib.contextmanager
    def edit_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_stylesheet()
        yield ss
        self.set_stylesheet(ss)

    def set_stylesheet(
        self, ss: None | str | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        match ss:
            case None:
                ss = ""
            case os.PathLike():
                ss = pathlib.Path(ss).read_text()
            case qstylizer.style.StyleSheet():
                ss = str(ss)
        self.setStyleSheet(ss)

    def get_stylesheet(self) -> qstylizer.style.StyleSheet:
        try:
            return qstylizer.parser.parse(self.styleSheet())
        except ValueError:
            return qstylizer.style.StyleSheet()

    @contextlib.contextmanager
    def edit_palette(self) -> Iterator[gui.Palette]:
        palette = gui.Palette(self.palette())
        yield palette
        self.setPalette(palette)

    @contextlib.contextmanager
    def edit_font(self) -> Iterator[gui.Font]:
        font = gui.Font(self.font())
        yield font
        self.setFont(font)

    def set_context_menu_policy(self, policy: constants.ContextPolicyStr) -> None:
        """Set contextmenu policy for given item view.

        Args:
            policy: contextmenu policy to use

        Raises:
            InvalidParamError: policy does not exist
        """
        if policy not in constants.CONTEXT_POLICY:
            raise InvalidParamError(policy, constants.CONTEXT_POLICY)
        self.setContextMenuPolicy(constants.CONTEXT_POLICY[policy])

    def get_context_menu_policy(self) -> constants.ContextPolicyStr:
        """Return current contextmenu policy.

        Returns:
            contextmenu policy
        """
        return constants.CONTEXT_POLICY.inverse[self.contextMenuPolicy()]

    def set_window_state(self, policy: constants.WindowStateStr) -> None:
        """Set window state for given item view.

        Args:
            policy: window state to use

        Raises:
            InvalidParamError: policy does not exist
        """
        if policy not in constants.WINDOW_STATES:
            raise InvalidParamError(policy, constants.WINDOW_STATES)
        self.setWindowState(constants.WINDOW_STATES[policy])

    def get_window_state(self) -> constants.WindowStateStr:
        """Return current window state.

        Returns:
            window state
        """
        return constants.WINDOW_STATES.inverse[self.windowState()]

    def set_custom_menu(self, method: Callable) -> None:
        self.set_context_menu_policy("custom")
        self.customContextMenuRequested.connect(method)

    def set_layout(
        self,
        layout: LayoutStr | QtWidgets.QLayout | None,
        margin: int | None = None,
        spacing: int | None = None,
    ) -> QtWidgets.QLayout:
        """Quick way to set a layout.

        Sets layout to given layout, also allows setting margin and spacing.

        Args:
            layout: Layout to set
            margin: margin to use in pixels
            spacing: spacing to use in pixels

        Returns:
            Layout
        """
        match layout:
            case None:
                return
            case "horizontal":
                self.box = widgets.HBoxLayout()
            case "vertical":
                self.box = widgets.VBoxLayout()
            case "grid":
                self.box = widgets.GridLayout()
            case "form":
                self.box = widgets.FormLayout()
            case "stacked":
                self.box = widgets.StackedLayout()
            case "flow":
                from prettyqt import custom_widgets

                self.box = custom_widgets.FlowLayout()
            case QtWidgets.QLayout():
                self.box = layout
            case _:
                raise ValueError("Invalid Layout")
        self.setLayout(self.box)
        if margin is not None:
            self.box.set_margin(margin)
        if spacing is not None:
            self.box.setSpacing(spacing)
        return self.box

    def position_on(
        self,
        where: Literal["parent", "window", "screen", "mouse"]
        | QtWidgets.QWidget
        | QtCore.QRect
        | QtCore.QPoint
        | tuple[int, int]
        | tuple[int, int, int, int],
        how: Literal[
            "center",
            "top",
            "left",
            "bottom",
            "right",
            "top_left",
            "top_right",
            "bottom_left",
            "bottom_right",
        ] = "center",
        scale_ratio: int | None = None,
        x_offset: int = 0,
        y_offset: int = 0,
    ):
        """Position widget on another widget / window / screen.

        Arguments:
            where: where to positin on
            how: How to align
            scale_ratio: Resize to scale_ratio * target size
            x_offset: additional x offset for final position
            y_offset: additional y offset for final position
        """
        do_scale = True
        match where:
            case "mouse":
                geom = core.Rect(gui.Cursor.pos(), gui.Cursor.pos())
                do_scale = False
            case QtCore.QPoint():
                geom = core.Rect(where, where)
                do_scale = False
            case (int(), int()):
                p = core.Point(*where)
                geom = core.Rect(p, p)
                do_scale = False
            case (int(), int(), int(), int()):
                geom = core.Rect(*where)
            case "parent":
                geom = self.parent().frameGeometry()
            case "window":
                geom = self.window().frameGeometry()
            case QtWidgets.QWidget():
                geom = where.frameGeometry()
            case QtCore.QRect():
                geom = where
            case "screen":
                geom = gui.GuiApplication.primaryScreen().geometry()
        if scale_ratio is not None and do_scale:
            self.resize(
                int(geom.width() * scale_ratio),
                int(geom.height() * scale_ratio),
            )
        own_geo = self.frameGeometry()
        match how:
            case "center":
                new = geom.center()
            case "top":
                new = core.Point(geom.center().x(), geom.top() + own_geo.height() // 2)
            case "bottom":
                new = core.Point(geom.center().x(), geom.bottom() - own_geo.height() // 2)
            case "left":
                new = core.Point(geom.left() + own_geo.width() // 2, geom.center().y())
            case "right":
                new = core.Point(geom.right() - own_geo.width() // 2, geom.center().y())
            case "top_right":
                new = core.Point(
                    geom.right() - own_geo.width() // 2,
                    geom.top() + own_geo.height() // 2,
                )
            case "top_left":
                new = core.Point(
                    geom.left() + own_geo.width() // 2,
                    geom.top() + own_geo.height() // 2,
                )
            case "bottom_right":
                new = core.Point(
                    geom.right() - own_geo.width() // 2,
                    geom.bottom() - own_geo.height() // 2,
                )
            case "bottom_left":
                new = core.Point(
                    geom.left() + own_geo.width() // 2,
                    geom.bottom() - own_geo.height() // 2,
                )
        new = core.Point(new.x() + x_offset, new.y() + y_offset)
        own_geo.moveCenter(new)
        self.move(own_geo.topLeft())

    def set_cursor(self, cursor: constants.CursorShapeStr | QtGui.QCursor) -> None:
        if isinstance(cursor, QtGui.QCursor):
            curs = cursor
        elif cursor in constants.CURSOR_SHAPE:
            curs = gui.Cursor(constants.CURSOR_SHAPE[cursor])
        else:
            raise InvalidParamError(cursor, constants.CURSOR_SHAPE)
        self.setCursor(curs)

    def set_focus_policy(self, policy: constants.FocusPolicyStr) -> None:
        """Set the way the widget accepts keyboard focus.

        Args:
            policy (str): Focus policy

        Raises:
            InvalidParamError: Description
        """
        if policy not in constants.FOCUS_POLICY:
            raise InvalidParamError(policy, constants.FOCUS_POLICY)
        self.setFocusPolicy(constants.FOCUS_POLICY[policy])

    def get_focus_policy(self) -> constants.FocusPolicyStr:
        """Return waay the widget accepts keyboard focus.

        Returns:
            str: Focus policy
        """
        return constants.FOCUS_POLICY.inverse[self.focusPolicy()]

    def set_focus(self, reason: constants.FocusReasonStr | None = None):
        if reason is None:
            self.setFocus()
        else:
            self.setFocus(constants.FOCUS_REASONS[reason])

    def set_font_size(self, size: int) -> None:
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)

    def get_font_metrics(self) -> gui.FontMetrics:
        return gui.FontMetrics(self.fontMetrics())

    def get_font_info(self) -> gui.FontInfo:
        return gui.FontInfo(self.fontInfo())

    def set_margin(self, margin: int) -> None:
        self.setContentsMargins(margin, margin, margin, margin)

    def raise_dock(self) -> bool:
        w = self.find_parent(QtWidgets.QDockWidget)
        if w is None:
            return False
        w.setVisible(True)
        w.raise_()
        return True

    def set_mask(
        self,
        area: datatypes.RectType | QtGui.QRegion | QtGui.QBitmap | None,
        typ: gui.region.RegionTypeStr = "rectangle",
    ):
        match area:
            case None:
                self.clearMask()
                return
            case tuple():
                area = gui.Region(*area, gui.region.REGION_TYPE[typ])
            case QtCore.QRect():
                area = gui.Region(area, gui.region.REGION_TYPE[typ])
        self.setMask(area)

    def set_window_file_path(self, path: datatypes.PathType):
        self.setWindowFilePath(os.fspath(path))

    def get_window_file_path(self) -> pathlib.Path | None:
        path = self.windowFilePath()
        return pathlib.Path(path) if path else None

    def get_image(self) -> QtGui.QPixmap:
        from prettyqt.qt import QtOpenGLWidgets

        image = self.grab()
        with gui.Painter(image) as painter:
            painter.set_composition_mode("source_atop")
            for gl_widget in self.find_children(QtOpenGLWidgets.QOpenGLWidget):
                d = gl_widget.mapToGlobal(core.Point()) - self.mapToGlobal(core.Point())
                painter.drawImage(d, gl_widget.grabFramebuffer())
        return image

    def get_screen(self) -> gui.Screen | None:
        window = self.window().windowHandle()
        return None if window is None else gui.Screen(window.screen())

    def delete_children(self):
        """Delete all children of the specified QObject."""
        if hasattr(self, "clear"):
            return self.clear()
        layout = self.layout()
        while layout.count():
            item = layout.takeAt(0)
            if widget := item.widget():
                widget.deleteLater()
            else:
                self.delete_children(item.layout())

    def get_cursor(self) -> gui.Cursor:
        return gui.Cursor(self.cursor())

    def set_style(self, style: str | QtWidgets.QStyle):
        if isinstance(style, str):
            style = QtWidgets.QStyleFactory.create(style)
        self.setStyle(style)

    def highlight_widget(self, widget):
        from prettyqt.custom_widgets.overlayborder import FocusWidget

        widget = FocusWidget(self, widget)
        widget.show()


class Widget(WidgetMixin, prettyprinter.PrettyPrinter, QtWidgets.QWidget):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.RadioButton("tsd")
    widget2 = Widget(status_tip="trekk", whats_this="kfjk", minimum_width=None)
    # widget.play_animation(
    #     "property",
    #     name="windowOpacity",
    #     duration=1000,
    #     start_value=0,
    #     end_value=1,
    # )
    # widget.play_animation("fade_in")
    # val = custom_animations.FadeInAnimation()
    # val.apply_to(widget)
    # val.start()
    widget.set_graphics_effect("colorize")
    widget.show()
    widget2.show()
    mainwindow = widgets.MainWindow()
    container = widgets.Widget()
    container.set_layout("horizontal")
    container.box.add(widget)
    container.box.add(widget2)
    mainwindow.setCentralWidget(container)
    mainwindow.show()
    mainwindow.highlight_widget(widget)
    app.sleep(4)
    # widget.position_on("mouse", scale_ratio=0.5, how="top")
    # widget.add_shortcut("return", print, "application")
    # widget.set_min_size((400, 400))
    # widget.set_max_size(None, 600)
    # print(type(widget.get_screen()))
    app.main_loop()



from __future__ import annotations

import datetime

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets


SortIndicator = QtWidgets.QStyleOptionHeader.SortIndicator
StateFlag = QtWidgets.QStyle.StateFlag
CE = QtWidgets.QStyle.ControlElement
SelectedPosition = QtWidgets.QStyleOptionHeader.SelectedPosition
SectionPosition = QtWidgets.QStyleOptionHeader.SectionPosition

HORIZONTAL_HEADER_DATA_ROLE = QtCore.Qt.ItemDataRole.UserRole + 150
VERTICAL_HEADER_DATA_ROLE = QtCore.Qt.ItemDataRole.UserRole + 151


class HierarchicalHeaderView(widgets.HeaderView):
    """Hierarchical header view.

    This class is a Python port of
    http://qt-apps.org/content/show.php/HierarchicalHeaderView?content=103154
    """

    class PrivateData:
        """Privatedata."""

        header_model = None

        def init_from_new_model(self, orientation: int, model: QtCore.QAbstractItemModel):
            is_hor = orientation == constants.HORIZONTAL
            role = HORIZONTAL_HEADER_DATA_ROLE if is_hor else VERTICAL_HEADER_DATA_ROLE
            self.header_model = model.data(core.ModelIndex(), role)

        def find_root_index(self, index: core.ModelIndex) -> core.ModelIndex:
            while index.parent().isValid():
                index = index.parent()
            return index

        def get_parent_indexes(self, index: core.ModelIndex) -> list:
            indexes = []
            while index.isValid():
                indexes.insert(0, index)
                index = index.parent()
            return indexes

        def find_leaf(
            self, index: core.ModelIndex, section_index: int, current_leaf_index: int
        ) -> core.ModelIndex:
            if not index.isValid():
                return core.ModelIndex(), current_leaf_index
            model = index.model()
            if child_count := model.columnCount(index):
                for i in range(child_count):
                    res, current_leaf_index = self.find_leaf(
                        model.index(0, i, index), section_index, current_leaf_index
                    )
                    if res.isValid():
                        return res, current_leaf_index
            else:
                current_leaf_index += 1
                if current_leaf_index == section_index:
                    return index, current_leaf_index
            return core.ModelIndex(), current_leaf_index

        def leaf_index(self, section_index: int) -> core.ModelIndex:
            if self.header_model:
                current_leaf_index = -1
                for i in range(self.header_model.columnCount()):
                    res, current_leaf_index = self.find_leaf(
                        self.header_model.index(0, i), section_index, current_leaf_index
                    )
                    if res.isValid():
                        return res
            return core.ModelIndex()

        def search_leafs(self, index: core.ModelIndex) -> list[core.ModelIndex]:
            res = []
            if index.isValid():
                model = index.model()
                if child_count := index.model().columnCount(index):
                    for i in range(child_count):
                        leaf = model.index(0, i, index)
                        res += self.search_leafs(leaf)
                else:
                    res.append(index)
            return res

        def leafs(self, index: core.ModelIndex) -> list[core.ModelIndex]:
            leafs = []
            if index.isValid():
                model = index.model()
                child_count = index.model().columnCount(index)
                for i in range(child_count):
                    leaf = model.index(0, i, index)
                    leafs += self.search_leafs(leaf)
            return leafs

        def set_foreground_brush(
            self, opt: QtWidgets.QStyleOptionHeader, index: core.ModelIndex
        ):
            if foreground_brush := index.data(constants.FOREGROUND_ROLE):
                brush = QtGui.QBrush(foreground_brush)
                opt.palette.setBrush(QtGui.QPalette.ColorRole.ButtonText, brush)

        def set_background_brush(
            self, opt: QtWidgets.QStyleOptionHeader, index: core.ModelIndex
        ):
            if background_brush := index.data(constants.BACKGROUND_ROLE):
                brush = QtGui.QBrush(background_brush)
                opt.palette.setBrush(QtGui.QPalette.ColorRole.Button, brush)
                opt.palette.setBrush(QtGui.QPalette.ColorRole.Window, brush)

        def get_cell_size(
            self,
            leaf_index: core.ModelIndex,
            hv: QtWidgets.QHeaderView,
            style_options: QtWidgets.QStyleOptionHeader,
        ) -> QtCore.QSize:
            res = QtCore.QSize()
            if variant := leaf_index.data(constants.SIZE_HINT_ROLE):
                res = variant
            fnt = var if (var := leaf_index.data(constants.FONT_ROLE)) else hv.font()
            fnt.setBold(True)
            fm = QtGui.QFontMetrics(fnt)
            text_size = fm.size(0, leaf_index.data(constants.DISPLAY_ROLE))
            size = text_size + QtCore.QSize(4, 0)
            if leaf_index.data(constants.USER_ROLE):
                size.transpose()
            decoration_size = hv.style().sizeFromContents(
                QtWidgets.QStyle.ContentsType.CT_HeaderSection,
                style_options,
                QtCore.QSize(),
                hv,
            )
            empty_text_size = fm.size(0, "")
            return res.expandedTo(size + decoration_size - empty_text_size)

        def get_current_cell_width(
            self,
            searched_index: core.ModelIndex,
            leaf_index: core.ModelIndex,
            section_index: int,
            hv: QtWidgets.QHeaderView,
        ) -> int:
            leafs_list = self.leafs(searched_index)
            if not leafs_list:
                return hv.sectionSize(section_index)
            offset = leafs_list.index(leaf_index) if leaf_index in leafs_list else -1
            first_leaf_section_index = section_index - offset
            return sum(
                hv.sectionSize(first_leaf_section_index + i)
                for i in range(len(leafs_list))
            )

        def get_current_cell_left(
            self,
            searched_index: core.ModelIndex,
            leaf_index: core.ModelIndex,
            section_index: int,
            left: int,
            hv: QtWidgets.QHeaderView,
        ) -> int:
            if leafs_list := self.leafs(searched_index):
                n = leafs_list.index(leaf_index) if leaf_index in leafs_list else -1
                first_leaf_section_index = section_index - n
                for i in range(n - 1, -1, -1):
                    left -= hv.sectionSize(first_leaf_section_index + i)
            return left

        def paint_horizontal_cell(
            self,
            painter: QtGui.QPainter,
            hv: QtWidgets.QHeaderView,
            cell_index: core.ModelIndex,
            leaf_index: core.ModelIndex,
            logical_leaf_index: int,
            style_options: QtWidgets.QStyleOptionHeader,
            section_rect: QtCore.QRect,
            top: int,
        ):
            uniopt = QtWidgets.QStyleOptionHeader(style_options)
            self.set_foreground_brush(uniopt, cell_index)
            self.set_background_brush(uniopt, cell_index)
            height = (
                section_rect.height() - top
                if cell_index == leaf_index
                else self.get_cell_size(cell_index, hv, uniopt).height()
            )
            left = self.get_current_cell_left(
                cell_index, leaf_index, logical_leaf_index, section_rect.left(), hv
            )
            width = self.get_current_cell_width(
                cell_index, leaf_index, logical_leaf_index, hv
            )
            r = QtCore.QRect(left, top, width, height)
            uniopt.text = cell_index.data(constants.DISPLAY_ROLE)
            painter.save()
            uniopt.rect = r
            style = hv.style()
            if cell_index.data(constants.USER_ROLE):
                style.drawControl(CE.CE_HeaderSection, uniopt, painter, hv)
                m = QtGui.QTransform()
                m.rotate(-90)
                painter.setWorldTransform(m, True)
                new_r = QtCore.QRect(0, 0, r.height(), r.width())
                new_r.moveCenter(QtCore.QPoint(-r.center().y(), r.center().x()))
                uniopt.rect = new_r
                style.drawControl(CE.CE_HeaderLabel, uniopt, painter, hv)
            else:
                style.drawControl(CE.CE_Header, uniopt, painter, hv)
            painter.restore()
            return top + height

        def paint_horizontal_section(
            self,
            painter: QtGui.QPainter,
            section_rect: QtCore.QRect,
            logical_leaf_index: int,
            hv: QtWidgets.QHeaderView,
            style_options: QtWidgets.QStyleOptionHeader,
            leaf_index: core.ModelIndex,
        ):
            #            print(logical_leaf_index)
            old_bo = painter.brushOrigin()
            top = section_rect.y()
            indexes = self.get_parent_indexes(leaf_index)
            for i, idx in enumerate(indexes):
                real_style_options = QtWidgets.QStyleOptionHeader(style_options)
                if i < len(indexes) - 1 and (
                    real_style_options.state & StateFlag.State_Sunken
                    or real_style_options.state & StateFlag.State_On
                ):
                    t = StateFlag.State_Sunken | StateFlag.State_On
                    real_style_options.state = real_style_options.state & ~t
                    # FIXME: parent items are not highlighted
                if i < len(indexes) - 1:  # Use sortIndicator for inner level only
                    real_style_options.sortIndicator = SortIndicator.None_
                #                if i==0:
                #                    print(self.leafs(indexes[i]), leaf_index)
                top = self.paint_horizontal_cell(
                    painter,
                    hv,
                    idx,
                    leaf_index,
                    logical_leaf_index,
                    real_style_options,
                    section_rect,
                    top,
                )
            painter.setBrushOrigin(old_bo)

        def paint_vertical_cell(
            self,
            painter: QtGui.QPainter,
            hv: QtWidgets.QHeaderView,
            cell_index: core.ModelIndex,
            leaf_index: core.ModelIndex,
            logical_leaf_index: int,
            style_options: QtWidgets.QStyleOptionHeader,
            section_rect: QtCore.QRect,
            left: int,
        ):
            uniopt = QtWidgets.QStyleOptionHeader(style_options)
            self.set_foreground_brush(uniopt, cell_index)
            self.set_background_brush(uniopt, cell_index)
            width = (
                section_rect.width() - left
                if cell_index == leaf_index
                else self.get_cell_size(cell_index, hv, uniopt).width()
            )
            top = self.get_current_cell_left(
                cell_index, leaf_index, logical_leaf_index, section_rect.top(), hv
            )
            height = self.get_current_cell_width(
                cell_index, leaf_index, logical_leaf_index, hv
            )
            r = QtCore.QRect(left, top, width, height)
            uniopt.text = cell_index.data(constants.DISPLAY_ROLE)
            painter.save()
            uniopt.rect = r
            style = hv.style()
            if cell_index.data(constants.USER_ROLE):
                style.drawControl(CE.CE_HeaderSection, uniopt, painter, hv)
                m = QtGui.QTransform()
                m.rotate(-90)
                painter.setWorldTransform(m, True)
                new_r = QtCore.QRect(0, 0, r.height(), r.width())
                new_r.moveCenter(QtCore.QPoint(-r.center().y(), r.center().x()))
                uniopt.rect = new_r
                style.drawControl(CE.CE_HeaderLabel, uniopt, painter, hv)
            else:
                style.drawControl(CE.CE_Header, uniopt, painter, hv)
            painter.restore()
            return left + width

        def paint_vertical_section(
            self,
            painter: QtGui.QPainter,
            section_rect: QtCore.QRect,
            logical_leaf_index: int,
            hv: QtWidgets.QHeaderView,
            style_options: QtWidgets.QStyleOptionHeader,
            leaf_index: core.ModelIndex,
        ):
            old_bo = painter.brushOrigin()
            left = section_rect.x()
            indexes = self.get_parent_indexes(leaf_index)
            for i, idx in enumerate(indexes):
                real_style_options = QtWidgets.QStyleOptionHeader(style_options)
                if i < len(indexes) - 1 and (
                    real_style_options.state & StateFlag.State_Sunken
                    or real_style_options.state & StateFlag.State_On
                ):
                    t = StateFlag.State_Sunken | StateFlag.State_On
                    real_style_options.state = real_style_options.state & ~t
                    # FIXME: parent items are not highlighted
                left = self.paint_vertical_cell(
                    painter,
                    hv,
                    idx,
                    leaf_index,
                    logical_leaf_index,
                    real_style_options,
                    section_rect,
                    left,
                )
            painter.setBrushOrigin(old_bo)

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation | constants.OrientationStr,
        parent: QtWidgets.QWidget,
    ):
        super().__init__(orientation, parent)
        self._pd = self.PrivateData()
        self.sectionResized.connect(self.on_section_resized)
        self.setHighlightSections(True)
        self.setSectionsClickable(True)
        self.show()  # force to be visible
        if orientation in {constants.HORIZONTAL, "horizontal"}:
            parent.setHorizontalHeader(self)
        else:
            parent.setVerticalHeader(self)
        self.sectionMoved.connect(self._on_section_moved)

    def _on_section_moved(self, logical_index, old_visual_index, new_visual_index):
        view = self.parent()
        model = view.model()
        if not hasattr(model, "reorder"):
            return  # reorder underlying data of models with /reorder/ def only
        if getattr(self, "manual_move", False):
            self.manual_move = False
            return
        self.manual_move = True
        self.moveSection(new_visual_index, old_visual_index)  # cancel move
        if not model.reorder(old_visual_index, new_visual_index, self.orientation()):
            return
        rng = sorted((old_visual_index, new_visual_index))

        if self.orientation() == constants.HORIZONTAL:
            options = [(view.columnWidth(i), i) for i in range(rng[0], rng[1] + 1)]
            for i, col in enumerate(range(rng[0], rng[1] + 1)):
                view.setColumnWidth(col, options[i][0])
            view.selectColumn(new_visual_index)
        else:
            options = [(view.rowHeight(i), i) for i in range(rng[0], rng[1] + 1)]
            for i, col in enumerate(range(rng[0], rng[1] + 1)):
                view.setRowHeight(col, options[i][0])
            view.selectRow(new_visual_index)

        # FIXME: don't select if sorting is enable?
        if self.isSortIndicatorShown():
            sort_ind_index = next(
                (i for i, o in enumerate(options) if o[1] == self.sortIndicatorSection()),
                None,
            )
            # sort indicator is among sections being reordered
            if sort_ind_index is not None:
                # FIXME: does unnecessary sorting
                self.setSortIndicator(sort_ind_index + rng[0], self.sortIndicatorOrder())
        model.layoutChanged.emit()  # update view

    def get_style_option_for_cell(
        self, logical_index: int
    ) -> QtWidgets.QStyleOptionHeader:
        opt = QtWidgets.QStyleOptionHeader()
        self.initStyleOption(opt)
        if self.isSortIndicatorShown() and self.sortIndicatorSection() == logical_index:
            asc = self.sortIndicatorOrder() == constants.ASCENDING
            opt.sortIndicator = SortIndicator.SortDown if asc else SortIndicator.SortUp
        if self.window().isActiveWindow():
            opt.state = opt.state | StateFlag.State_Active
        opt.textAlignment = QtCore.Qt.AlignmentFlag.AlignCenter
        opt.iconAlignment = QtCore.Qt.AlignmentFlag.AlignVCenter
        opt.section = logical_index
        visual = self.visualIndex(logical_index)
        if self.count() == 1:
            opt.position = SectionPosition.OnlyOneSection
        elif visual == 0:
            opt.position = SectionPosition.Beginning
        else:
            is_end = visual == self.count() - 1
            opt.position = SectionPosition.End if is_end else SectionPosition.Middle
        sel_model = self.selectionModel()
        if not sel_model:
            return opt
        root_idx = self.rootIndex()
        if self.sectionsClickable() and self.highlightSections():
            if self.orientation() == constants.HORIZONTAL:
                if sel_model.columnIntersectsSelection(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_On
                if sel_model.isColumnSelected(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_Sunken
            else:
                if sel_model.rowIntersectsSelection(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_On
                if sel_model.isRowSelected(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_Sunken
        prev_idx = self.logicalIndex(visual - 1)
        next_idx = self.logicalIndex(visual + 1)
        if self.orientation() == constants.HORIZONTAL:
            prev_selected = sel_model.isColumnSelected(prev_idx, root_idx)
            next_selected = sel_model.isColumnSelected(next_idx, root_idx)
        else:
            prev_selected = sel_model.isRowSelected(prev_idx, root_idx)
            next_selected = sel_model.isRowSelected(next_idx, root_idx)

        if prev_selected and next_selected:
            opt.selectedPosition = SelectedPosition.NextAndPreviousAreSelected
        elif prev_selected:
            opt.selectedPosition = SelectedPosition.PreviousIsSelected
        elif next_selected:
            opt.selectedPosition = SelectedPosition.NextIsSelected
        else:
            opt.selectedPosition = SelectedPosition.NotAdjacent
        return opt

    def sectionSizeFromContents(self, logical_index: int) -> QtCore.QSize:
        if not self._pd.header_model:
            return super().sectionSizeFromContents(logical_index)
        cur_leaf_index = self._pd.leaf_index(logical_index)
        if not cur_leaf_index.isValid():
            return super().sectionSizeFromContents(logical_index)
        styleOption = QtWidgets.QStyleOptionHeader(
            self.get_style_option_for_cell(logical_index)
        )
        s = self._pd.get_cell_size(cur_leaf_index, self, styleOption)
        cur_leaf_index = cur_leaf_index.parent()
        while cur_leaf_index.isValid():
            cell_size = self._pd.get_cell_size(cur_leaf_index, self, styleOption)
            if self.orientation() == constants.HORIZONTAL:
                s.setHeight(s.height() + cell_size.height())
            else:
                s.setWidth(s.width() + cell_size.width())
            cur_leaf_index = cur_leaf_index.parent()
        return s

    def paintSection(
        self, painter: QtGui.QPainter, rect: QtCore.QRect, logical_index: int
    ):
        if not rect.isValid():
            super().paintSection(painter, rect, logical_index)
            return
        leaf_index = self._pd.leaf_index(logical_index)
        if not leaf_index.isValid():
            super().paintSection(painter, rect, logical_index)
            return
        style_option = self.get_style_option_for_cell(logical_index)
        if self.orientation() == constants.HORIZONTAL:
            self._pd.paint_horizontal_section(
                painter,
                rect,
                logical_index,
                self,
                style_option,
                leaf_index,
            )
        else:
            self._pd.paint_vertical_section(
                painter,
                rect,
                logical_index,
                self,
                style_option,
                leaf_index,
            )

    def on_section_resized(self, logical_index: int):
        if self.isSectionHidden(logical_index):
            return
        leaf_index = self._pd.leaf_index(logical_index)
        if leaf_index.isValid():
            leafs_list = self._pd.leafs(self._pd.find_root_index(leaf_index))
            start = leafs_list.index(leaf_index) if leaf_index in leafs_list else -1
            is_horizontal = self.orientation() == constants.HORIZONTAL
            for _ in range(start, 0, -1):
                logical_index -= 1
                w = self.viewport().width()
                h = self.viewport().height()
                pos = self.sectionViewportPosition(logical_index)
                r = QtCore.QRect(pos, 0, w - pos, h)
                if is_horizontal:
                    if self.isRightToLeft():
                        r.setRect(0, 0, pos + self.sectionSize(logical_index), h)
                else:
                    r.setRect(0, pos, w, h - pos)
                self.viewport().update(r.normalized())

    def setModel(self, model: QtCore.QAbstractItemModel):
        super().setModel(model)
        model.layoutChanged.connect(self.layoutChanged)
        self.layoutChanged()

    def layoutChanged(self):
        if model := self.model():
            self._pd.init_from_new_model(self.orientation(), model)
            is_horizontal = self.orientation() == constants.HORIZONTAL
            count = model.columnCount() if is_horizontal else model.rowCount()
            self.initializeSections(0, count - 1)


class DataFrameModel(QtCore.QAbstractTableModel):
    # na_values:least|greatest - for sorting
    options = {
        "na_values": "least",
        "tooltip_min_len": 21,
    }

    def __init__(self, dataframe=None):
        super().__init__()
        self.set_dataframe(dataframe if dataframe is not None else pd.DataFrame())

    def set_dataframe(self, dataframe):
        self.df = dataframe.copy()
        self.layoutChanged.emit()

    def rowCount(self, parent=None):
        return len(self.df)

    def columnCount(self, parent=None):
        return len(self.df.columns)

    def read_level(self, y: int = 0, xs: int = 0, xe=None, orient=None):
        c = self.df.columns if orient == HORIZONTAL_HEADER_DATA_ROLE else self.df.index
        if not hasattr(c, "levels"):  # not MultiIndex
            return [gui.StandardItem(str(i)) for i in c]
        sibl = []
        section_start, v, xe = xs, None, xe or len(c)
        for i in range(xs, xe):
            label = c.codes[y][i]
            if label != v:
                if y + 1 < len(c.levels) and i > xs:
                    children = self.read_level(y + 1, section_start, i, orient=orient)
                    sibl[-1].appendRow(children)
                item = gui.StandardItem(str(c.levels[y][label]))
                sibl.append(item)
                section_start = i
                v = label
        if y + 1 < len(c.levels):
            children = self.read_level(y + 1, section_start, orient=orient)
            sibl[-1].appendRow(children)
        return sibl

    def data(self, index, role=constants.DISPLAY_ROLE):
        row, col = index.row(), index.column()
        if role in (constants.DISPLAY_ROLE, constants.TOOLTIP_ROLE):
            ret = self.df.iat[row, col]
            if ret is not None:  # convert to str except for None, NaN, NaT
                if isinstance(ret, float):
                    ret = f"{ret:n}"
                elif isinstance(ret, datetime.date):
                    # FIXME: show microseconds optionally
                    ret = ret.strftime(
                        "%c" if isinstance(ret, datetime.datetime) else "%x"
                    )
                else:
                    ret = str(ret)
                if (
                    role == constants.TOOLTIP_ROLE
                    and len(ret) < self.options["tooltip_min_len"]
                ):
                    ret = ""
                return ret
        elif role in (HORIZONTAL_HEADER_DATA_ROLE, VERTICAL_HEADER_DATA_ROLE):
            hm = gui.StandardItemModel()
            hm.appendRow(self.read_level(orient=role))
            return hm

    def reorder(self, old_index, new_index, orientation):
        """Reorder columns / rows."""
        horizontal = orientation == constants.HORIZONTAL
        cols = list(self.df.columns if horizontal else self.df.index)
        cols.insert(new_index, cols.pop(old_index))
        self.df = self.df[cols] if horizontal else self.df.T[cols].T
        return True

    #    def filter(self, filt=None):
    #        self.df = self.df_full if filt is None else self.df[filt]
    #        self.layoutChanged.emit()

    def headerData(self, section, orientation, role):
        if role != constants.DISPLAY_ROLE:
            return
        if orientation == constants.HORIZONTAL:
            idx = self.df.columns
            return "\n".join(str(i) for i in idx) if isinstance(idx, tuple) else str(idx)
        else:
            idx = self.df.index
            return " | ".join(str(i) for i in idx) if isinstance(idx, tuple) else str(idx)
        # return label if type(label) is tuple else label

    def sort(self, column, order):
        print("sort", column, order)
        if len(self.df):
            asc = order == constants.ASCENDING
            na_pos = "first" if (self.options["na_values"] == "least") == asc else "last"
            col = self.df.columns[column]
            self.df.sort_values(col, ascending=asc, inplace=True, na_position=na_pos)
            self.layoutChanged.emit()


if __name__ == "__main__":
    import locale
    import sys

    import numpy as np
    import pandas as pd

    from prettyqt import gui

    locale.setlocale(locale.LC_ALL, "")  # system locale settings
    app = QtWidgets.QApplication(sys.argv)
    form = QtWidgets.QWidget()
    # form.setAttribute(Qt.WA_DeleteOnClose)  # http://stackoverflow.com/a/27178019/1119602
    form.setMinimumSize(700, 260)
    view = widgets.TableView()
    widgets.VBoxLayout(form).addWidget(view)
    form.show()

    # Prepare data
    tuples = [
        ("bar", "one", "q"),
        ("bar", "two", "q"),
        ("baz", "one", "q"),
        ("baz", "two", "q"),
        ("foo", "one", "q"),
        ("foo", "two", "q"),
        ("qux", "one", "q"),
        ("qux", "two", "q"),
    ]
    index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
    df = pd.DataFrame(np.random.randn(6, 6), index=index[:6], columns=index[:6])
    print("DataFrame:\n%s" % df)

    # Prepare view
    #    oldh, oldv = view.horizontalHeader(), view.verticalHeader()
    #    oldh.setParent(form), oldv.setParent(form) #Save old headers for some reason
    HierarchicalHeaderView(constants.HORIZONTAL, view)
    HierarchicalHeaderView(constants.VERTICAL, view)
    view.h_header.setSectionsMovable(True)  # reorder DataFrame columns manually
    # Set data
    model = DataFrameModel(df)
    view.setModel(model)
    idx = model.index(0, 0)
    idx.multiData(33)
    view.set_sorting_enabled(True, False)
    view.resizeColumnsToContents()
    view.resizeRowsToContents()
    # Set sorting enabled (after setting model)
    sys.exit(app.exec())



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.utils import datatypes


class CursorMoveAnimation(core.VariantAnimation):
    def __init__(
        self,
        duration: int = 1000,
        end_value: datatypes.PointType = (0, 0),
        start_value: datatypes.PointType | None = None,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.set_easing(easing)

        start = gui.Cursor.pos() if start_value is None else core.Point(*start_value)
        self.set_start_value(start)
        self.set_end_value(end_value)
        self.setDuration(duration)
        self.valueChanged.connect(gui.Cursor.setPos)

    def set_start_value(self, point: datatypes.PointType):
        if isinstance(point, tuple):
            point = core.Point(*point)
        self.setStartValue(point)

    def set_end_value(self, point: datatypes.PointType):
        if isinstance(point, tuple):
            point = core.Point(*point)
        self.setEndValue(point)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    animation = CursorMoveAnimation()
    animation.set_end_value((100, 100))
    btn = widgets.PushButton("ts")
    btn.show()
    animation.start()
    app.main_loop()



from __future__ import annotations

import sys

from prettyqt import constants, core, custom_models, iconprovider, widgets
from prettyqt.qt import QtGui


# TODO: Set icon colour and copy code with color kwarg

ALL_COLLECTIONS = "All"


class IconBrowser(widgets.MainWindow):
    """A small browser window that allows the user to search through all icons.

    You can also copy the name and python code for the currently selected icon.
    """

    def __init__(self):
        super().__init__()
        self.setMinimumSize(500, 500)
        self.set_title("Icon Browser")
        font_maps = {k: v.charmap for k, v in iconprovider._instance().fonts.items()}
        icon_names = [
            f"{font_collection}.{icon_name}"
            for font_collection, font_data in font_maps.items()
            for icon_name in font_data
        ]
        model = IconModel(self.get_palette().get_color("text"))
        model.setStringList(sorted(icon_names))

        self._proxy_model = custom_models.FuzzyFilterProxyModel()
        self._proxy_model.setSourceModel(model)
        self._proxy_model.set_filter_case_sensitive(True)
        self._proxy_model.set_match_color(None)

        self._listview = IconListView(
            self,
            uniform_item_sizes=True,
            view_mode="icon",
            context_menu_policy="custom",
        )

        self._listview.set_model(self._proxy_model)
        self._listview.doubleClicked.connect(self._copy_icon_text)

        self._lineedit = widgets.LineEdit(parent=self)
        self._lineedit.textChanged.connect(self._trigger_instant_update)

        self._combobox = widgets.ComboBox(parent=self)
        self._combobox.setMinimumWidth(75)
        self._combobox.currentIndexChanged.connect(self._trigger_instant_update)
        self._combobox.addItems([ALL_COLLECTIONS, *sorted(font_maps.keys())])

        search_bar_frame = widgets.Frame(self)
        with widgets.HBoxLayout.create(search_bar_frame, margin=0) as layout:
            layout.add(self._combobox)
            layout.add(self._lineedit)

        self._copy_button = widgets.PushButton("Copy Name", clicked=self._copy_icon_text)
        frame = widgets.Frame(self)
        with widgets.VBoxLayout.create(frame) as layout:
            layout.add(search_bar_frame)
            layout.add(self._listview)
            layout.add(self._copy_button)
        self.setCentralWidget(frame)
        self.add_shortcut("return", self._copy_icon_text)
        self._lineedit.setFocus()
        self.position_on("screen")

    def _trigger_instant_update(self):
        """Stop timer used for committing search term and update proxy model instantly."""
        self._proxy_model.set_search_term(self._lineedit.text())

    def _copy_icon_text(self):
        """Copy the name of the currently selected icon to the clipboard."""
        if indexes := self._listview.selectedIndexes():
            widgets.Application.copy_to_clipboard(indexes[0].data())


class IconListView(widgets.ListView):
    """A QListView that scales its grid size to always show same amount of items."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_vertical_scrollbar_policy("always_on")
        self.VIEW_COLUMNS = 5

    def resizeEvent(self, event):
        """Re-calculate the grid size to provide scaling icons."""
        width = self.viewport().width() - 30
        # The minus 30 above ensures we don't end up with an item width that
        # can't be drawn the expected number of times across the view without
        # being wrapped. Without this, the view can flicker during resize
        tile_width = int(width / self.VIEW_COLUMNS)
        icon_width = int(tile_width * 0.8)
        self.set_grid_size((tile_width, tile_width))
        self.set_icon_size((icon_width, icon_width))

        return super().resizeEvent(event)


class IconModel(core.StringListModel):
    def __init__(self, icon_color: QtGui.QColor):
        super().__init__()
        self._icon_color = icon_color

    def flags(self, index):
        return constants.IS_ENABLED | constants.IS_SELECTABLE  # type: ignore

    def data(self, index, role=constants.DISPLAY_ROLE):
        if role == constants.DECORATION_ROLE:
            icon_string = self.data(index, role=constants.DISPLAY_ROLE)
            return iconprovider._icon(icon_string, color=self._icon_color)
        return super().data(index, role)


def run():
    app = widgets.app()
    browser = IconBrowser()
    browser.show()
    sys.exit(app.main_loop())


if __name__ == "__main__":
    run()



from __future__ import annotations

import re

from prettyqt import core, gui
from prettyqt.qt import QtGui


class SelectedWordHighlighter(gui.SyntaxHighlighter):
    def __init__(self, parent: QtGui.QTextDocument | None = None):
        super().__init__(parent)
        self.selection_term = ""
        self._highlight_format = gui.TextCharFormat()
        self._highlight_format.setBackground(gui.Color(255, 210, 120))
        self._highlight_format.setFontWeight(QtGui.QFont.Weight.Bold)
        self._highlight_pattern = None
        self._widget = parent.parent()
        self._widget.selectionChanged.connect(self.filter_selection_for_single_word)

    def set_selection_term(self, term: str):
        if term == self.selection_term:
            return
        if term:
            term = f"\\b{term}\\b"
            if term == self.selection_term:
                return
        self.selection_term = term
        self._highlight_pattern = core.RegularExpression(self.selection_term)
        self.rehighlight()

    def highlightBlock(self, text):
        if len(self.selection_term) > 1:
            self.apply_highlight(text)

    def apply_highlight(self, text):
        for m in self._highlight_pattern.finditer(text):
            length = m.span()[1] - m.span()[0]
            self.setFormat(m.span()[0], length, self._highlight_format)

    def filter_selection_for_single_word(self):
        tc = self._widget.get_text_cursor()
        current_selection = tc.selectedText()
        pat = re.compile("\\s+")
        items = pat.split(current_selection)  # SkipEmptyParts
        if not items:
            self.set_selection_term("")
            return
        tc.select_text("start_of_word", "end_of_word")
        term = current_selection if current_selection == tc.selectedText() else ""
        self.set_selection_term(term)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    editor = widgets.PlainTextEdit()
    highlighter = SelectedWordHighlighter(editor.document())
    editor.set_syntaxhighlighter(highlighter)
    # highlighter = SelectedWordHighlighter(editor.document())
    editor.show()
    app.main_loop()



from __future__ import annotations

import enum
from typing import Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import InvalidParamError, bidict, colors, datatypes


SCENE_LAYER = bidict(
    item=QtWidgets.QGraphicsScene.SceneLayer.ItemLayer,
    background=QtWidgets.QGraphicsScene.SceneLayer.BackgroundLayer,
    foreground=QtWidgets.QGraphicsScene.SceneLayer.ForegroundLayer,
    all=QtWidgets.QGraphicsScene.SceneLayer.AllLayers,
)

SceneLayerStr = Literal["item", "background", "foreground", "all"]

ITEM_INDEX_METHOD = bidict(
    bsp_tree=QtWidgets.QGraphicsScene.ItemIndexMethod.BspTreeIndex,
    none=QtWidgets.QGraphicsScene.ItemIndexMethod.NoIndex,
)

ItemIndexMethodStr = Literal["bsp_tree", "none"]


class GraphicsScene(core.ObjectMixin, QtWidgets.QGraphicsScene):
    class GridType(enum.IntEnum):
        """Grid type for background."""

        NoGrid = 0
        DotGrid = 1
        LineGrid = 2

    def __init__(self, parent=None):
        super().__init__(parent)
        self._grid_mode = self.GridType.LineGrid
        self._grid_size = 50
        self._pen_width = 0.65
        self._grid_color = self.get_palette().get_color("text")
        self._bg_color = self.get_palette().get_color("window")
        # self.setBackgroundBrush(self._bg_color)

    def __repr__(self):
        cls_name = str(self.__class__.__name__)
        return f'<{cls_name}("{self.viewer()}") object at {hex(id(self))}>'

    def __getitem__(self, index: int) -> QtWidgets.QGraphicsItem:
        return self.items()[index]

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"itemIndexMethod": ITEM_INDEX_METHOD}
        return maps

    def get_palette(self) -> gui.Palette:
        return gui.Palette(self.palette())

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_background_brush(self) -> gui.Brush:
        return gui.Brush(self.backgroundBrush())

    def get_foreground_brush(self) -> gui.Brush:
        return gui.Brush(self.foregroundBrush())

    def add(self, item) -> QtWidgets.QGraphicsItem:
        match item:
            case QtWidgets.QGraphicsItem():
                self.addItem(item)
                return item
            case QtGui.QPixmap():
                return self.add_pixmap(item)
            case QtGui.QPainterPath():
                return self.add_path(item)
            case QtGui.QPolygonF():
                return self.add_polygon(item)
            case QtCore.QRectF():
                return self.add_rect(item)
            case QtCore.QLine():
                return self.add_line(item)
            case str():
                return self.add_text(item)
            case QtWidgets.QWidget():
                return self.add_widget(item)
            case _:
                raise TypeError(item)

    def add_pixmap(self, pixmap: QtGui.QPixmap) -> widgets.GraphicsPixmapItem:
        g_item = widgets.GraphicsPixmapItem(pixmap)
        self.addItem(g_item)
        return g_item

    def add_polygon(
        self,
        polygon: QtGui.QPolygonF | QtGui.QPolygon,
        pen: QtGui.QPen | None = None,
        brush: QtGui.QBrush | None = None,
    ) -> widgets.GraphicsPolygonItem:
        if isinstance(polygon, QtGui.QPolygon):
            polygon = gui.PolygonF(polygon)
        g_item = widgets.GraphicsPolygonItem(polygon)
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_path(
        self,
        path: QtGui.QPainterPath,
        pen: QtGui.QPen | None = None,
        brush: QtGui.QBrush | None = None,
    ) -> widgets.GraphicsPathItem:
        g_item = widgets.GraphicsPathItem(path)
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_rect(
        self,
        rect: datatypes.RectType | datatypes.RectFType,
        pen: QtGui.QPen | None = None,
        brush: QtGui.QBrush | None = None,
    ) -> widgets.GraphicsRectItem:
        if isinstance(rect, QtCore.QRect):
            rect = core.RectF(rect)
        elif isinstance(rect, tuple):
            rect = core.RectF(*rect)
        g_item = widgets.GraphicsRectItem(rect)
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_line(
        self,
        line: QtCore.QLineF | QtCore.QLine | tuple[float, float, float, float],
        pen: QtGui.QPen | None = None,
    ) -> widgets.GraphicsLineItem:
        if isinstance(line, QtCore.QLine):
            line = core.LineF(line)
        elif isinstance(line, tuple):
            line = core.LineF(*line)
        g_item = widgets.GraphicsLineItem(line)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_ellipse(
        self,
        ellipse: datatypes.RectType | datatypes.RectFType,
        pen: QtGui.QPen | None = None,
        brush: QtGui.QBrush | None = None,
    ) -> widgets.GraphicsEllipseItem:
        if isinstance(ellipse, QtCore.QRect):
            ellipse = core.RectF(ellipse)
        elif isinstance(ellipse, tuple):
            ellipse = core.RectF(*ellipse)
        g_item = widgets.GraphicsEllipseItem(ellipse)
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_text(
        self, text: str, font: QtGui.QFont | None = None
    ) -> widgets.GraphicsTextItem:
        g_item = widgets.GraphicsTextItem(text)
        if font is not None:
            g_item.setFont(font)
        self.addItem(g_item)
        return g_item

    def add_simple_text(
        self, text: str, font: QtGui.QFont | None = None
    ) -> widgets.GraphicsSimpleTextItem:
        g_item = widgets.GraphicsSimpleTextItem(text)
        if font is not None:
            g_item.setFont(font)
        self.addItem(g_item)
        return g_item

    def add_widget(self, widget: QtWidgets.QWidget) -> widgets.GraphicsProxyWidget:
        g_item = widgets.GraphicsProxyWidget()
        g_item.setWidget(widget)
        self.addItem(g_item)
        return g_item

    def colliding_items(
        self,
        item: QtWidgets.QGraphicsItem,
        mode: constants.ItemSelectionModeStr = "intersects_shape",
    ) -> list[QtWidgets.QGraphicsItem]:
        if mode not in constants.ITEM_SELECTION_MODE:
            raise InvalidParamError(mode, constants.ITEM_SELECTION_MODE)
        return self.collidingItems(item, constants.ITEM_SELECTION_MODE[mode])

    def add_item_group(
        self, *items: QtWidgets.QGraphicsItem
    ) -> widgets.GraphicsItemGroup:
        group = widgets.GraphicsItemGroup()
        for item in items:
            group.addToGroup(item)
        return group

    def _draw_grid(
        self,
        painter: QtGui.QPainter,
        rect: QtCore.QRectF,
        pen: QtGui.QPen,
        grid_size: int,
    ):
        left = int(rect.left())
        right = int(rect.right())
        top = int(rect.top())
        bottom = int(rect.bottom())

        first_left = left - (left % grid_size)
        first_top = top - (top % grid_size)

        lines = [
            QtCore.QLineF(x, top, x, bottom) for x in range(first_left, right, grid_size)
        ]
        lines.extend(
            [
                QtCore.QLineF(left, y, right, y)
                for y in range(first_top, bottom, grid_size)
            ]
        )

        painter.setPen(pen)
        painter.drawLines(lines)

    def _draw_dots(
        self,
        painter: QtGui.QPainter,
        rect: QtCore.QRectF,
        pen: QtGui.QPen,
        grid_size: int,
    ):
        if (zoom := self._get_viewer_zoom()) < 0:
            grid_size *= int(abs(zoom) / 0.3 + 1)

        left = int(rect.left())
        right = int(rect.right())
        top = int(rect.top())
        bottom = int(rect.bottom())

        first_left = left - (left % grid_size)
        first_top = top - (top % grid_size)

        pen.setWidth(grid_size / 10)
        painter.setPen(pen)

        [
            painter.drawPoint(int(x), int(y))
            for x in range(first_left, right, grid_size)
            for y in range(first_top, bottom, grid_size)
        ]

    def drawBackground(self, painter: QtGui.QPainter, rect: QtCore.QRect):
        super().drawBackground(painter, rect)

        painter.save()
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, False)
        painter.setBrush(self.backgroundBrush())

        if self._grid_mode == self.GridType.DotGrid:
            pen = QtGui.QPen(self.grid_color, self._pen_width)
            self._draw_dots(painter, rect, pen, self._grid_size)

        elif self._grid_mode == self.GridType.LineGrid:
            zoom = self._get_viewer_zoom()
            if zoom > -0.5:
                pen = QtGui.QPen(self.grid_color, self._pen_width)
                self._draw_grid(painter, rect, pen, self._grid_size)
            color = self._bg_color.darker(150)
            if zoom < -0.0:
                color = color.darker(100 - int(zoom * 110))
            pen = QtGui.QPen(color, self._pen_width)
            self._draw_grid(painter, rect, pen, self._grid_size * 8)

        painter.restore()

    def _get_viewer_zoom(self):
        transform = self.viewer().transform()
        cur_scale = (transform.m11(), transform.m22())
        return float(f"{cur_scale[0] - 1.0:0.2f}")

    def mousePressEvent(self, event):
        selected = self.viewer().selectedItems()
        if viewer := self.viewer():
            viewer.sceneMousePressEvent(event)
        super().mousePressEvent(event)
        keep_selection = any(
            [
                event.button() == QtCore.Qt.MiddleButton,
                event.button() == QtCore.Qt.RightButton,
                event.modifiers() == QtCore.Qt.AltModifier,
            ]
        )
        if keep_selection:
            for node in selected:
                node.setSelected(True)

    def mouseMoveEvent(self, event):
        if viewer := self.viewer():
            viewer.sceneMouseMoveEvent(event)
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if viewer := self.viewer():
            viewer.sceneMouseReleaseEvent(event)
        super().mouseReleaseEvent(event)

    def viewer(self):
        return self.views()[0] if self.views() else None

    def get_grid_mode(self) -> GridType:
        return self._grid_mode

    def set_grid_mode(self, mode: GridType | None = None):
        # alternative?
        # brush = gui.Brush()
        # brush.set_style("cross")
        # scene.setBackgroundBrush(brush)
        if mode is None:
            mode = self.GridType.NoGrid
        self._grid_mode = mode

    def get_grid_color(self) -> gui.Color:
        return self._grid_color

    def set_grid_color(self, color: datatypes.ColorType):
        self._grid_color = colors.get_color(color)

    def get_background_color(self) -> gui.Color:
        return self._bg_color

    def set_background_color(self, color: datatypes.ColorType):
        self._bg_color = colors.get_color(color)
        self.setBackgroundBrush(self._bg_color)

    def set_item_index_method(self, method: ItemIndexMethodStr):
        """Set item index method.

        Args:
            method: item index method to use

        Raises:
            InvalidParamError: invalid item index method
        """
        if method not in ITEM_INDEX_METHOD:
            raise InvalidParamError(method, ITEM_INDEX_METHOD)
        self.setItemIndexMethod(ITEM_INDEX_METHOD[method])

    def get_item_index_method(self) -> ItemIndexMethodStr:
        """Return item index method.

        Returns:
            item index method
        """
        return ITEM_INDEX_METHOD.inverse[self.itemIndexMethod()]

    bg_color = core.Property(QtGui.QColor, get_background_color, set_background_color)
    grid_color = core.Property(QtGui.QColor, get_grid_color, set_grid_color)
    grid_mode = core.Property(int, get_grid_mode, set_grid_mode)


if __name__ == "__main__":
    app = widgets.app()
    scene = GraphicsScene()
    scene.add_line(core.Line(0, 0, 10, 10))
    view = widgets.GraphicsView(scene)
    view.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import constants, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError


class TreeWidgetMixin(widgets.TreeViewMixin):
    def __contains__(self, other: QtWidgets.QTreeWidgetItem):
        return self.indexOfTopLevelItem(other) >= 0

    def sort(self, column: int = 0, reverse: bool = False):
        order = constants.DESCENDING if reverse else constants.ASCENDING
        self.sortItems(column, order)

    def find_items(
        self,
        text: str,
        column: int = 0,
        mode: constants.MatchFlagStr = "exact",
        recursive: bool = False,
        case_sensitive: bool = False,
    ) -> list[QtWidgets.QTreeWidgetItem]:
        if mode not in constants.MATCH_FLAGS:
            raise InvalidParamError(mode, constants.MATCH_FLAGS)
        flag = constants.MATCH_FLAGS[mode]
        if recursive:
            flag |= QtCore.Qt.MatchFlag.MatchRecursive
        if case_sensitive:
            flag |= QtCore.Qt.MatchFlag.MatchCaseSensitive
        return self.findItems(text, flag, column)  # type: ignore

    def get_items(self, recursive: bool = False) -> list[QtWidgets.QTreeWidgetItem]:
        """Get TreeWidgetItems of this widget.

        Arguments:
            recursive: whether to include all items of the tree.

        recursive option is written iteratively to also support original QTreeWidgetItems.
        """
        root = self.invisibleRootItem(())
        if not recursive:
            return [root.child(i) for i in range(root.childCount())]
        results = []
        nodes = [root]
        while nodes:
            items = []
            for node in nodes:
                results.append(node)
                items.extend(node.child(i) for i in range(node.childCount()))
            nodes = items
        return results[1:]

    def scroll_to_item(
        self,
        item: QtWidgets.QTreeWidgetItem,
        scroll_hint: widgets.abstractitemview.ScrollHintStr = "ensure_visible",
    ):
        self.scrollToItem(item, widgets.abstractitemview.SCROLL_HINT[scroll_hint])

    def removeTopLevelItem(self, item):
        for i in range(self.topLevelItemCount()):
            if self.topLevelItem(i) is item:
                self.takeTopLevelItem(i)
                return
        raise IndexError(f"Item '{str(item)}' not in top-level items.")

    def openPersistentEditor(
        self, index: QtCore.QModelIndex | QtWidgets.QTreeWidgetItem, column: int = 0
    ):
        if isinstance(index, QtCore.QModelIndex):
            index = self.itemFromIndex(index)
            column = index.column()
        super().openPersistentEditor(index, column)

    def closePersistentEditor(
        self, index: QtCore.QModelIndex | QtWidgets.QTreeWidgetItem, column: int = 0
    ):
        if isinstance(index, QtCore.QModelIndex):
            index = self.itemFromIndex(index)
            column = index.column()
        super().closePersistentEditor(index, column)

    def isPersistentEditorOpen(
        self, index: QtCore.QModelIndex | QtWidgets.QTreeWidgetItem, column: int = 0
    ) -> bool:
        if isinstance(index, QtCore.QModelIndex):
            index = self.itemFromIndex(index)
            column = index.column()
        return super().isPersistentEditorOpen(index, column)


class TreeWidget(TreeWidgetMixin, QtWidgets.QTreeWidget):
    pass


if __name__ == "__main__":
    from prettyqt.qt import QtCore

    app = widgets.app()
    widget = TreeWidget()
    widget.openPersistentEditor(QtCore.QModelIndex(), 1)
    widget.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import Any

from prettyqt import core, custom_models
from prettyqt.utils import treeitem


COL_NAME = custom_models.ColumnItem(
    name="Name",
    doc="Name.",
    label=lambda x: x.obj.key,
)

COL_VALUE = custom_models.ColumnItem(
    name="Value",
    doc="Value.",
    label=lambda x: repr(x.obj.value),
)

COL_TYPE = custom_models.ColumnItem(
    name="Type",
    doc="Type.",
    label=lambda x: repr(x.obj.typ),
)

COLUMNS = [COL_NAME, COL_VALUE, COL_TYPE]


@dataclass
class JsonItem:
    key: str | int
    value: Any
    typ: type | None = None


class JsonModel(custom_models.ColumnItemModel):
    """Model that provides an interface to an objectree that is build of tree items."""

    def __init__(
        self,
        obj: Any,
        show_root: bool = True,
        **kwargs,
    ):
        super().__init__(
            obj=JsonItem(key="", value=obj, typ=type(obj)),
            columns=COLUMNS,
            show_root=show_root,
            **kwargs,
        )

    def hasChildren(self, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        if parent.column() > 0:
            return False
        item = self.data_by_index(parent)
        if self.show_root and item == self._root_item:
            return True
        return isinstance(item.obj.value, dict | list | set)

    def _fetch_object_children(self, item: treeitem.TreeItem) -> list[treeitem.TreeItem]:
        """Fetch the children of a Python object.

        Returns: list of treeitem.TreeItems
        """
        # items = []
        match item.obj.value:
            case Mapping():
                return [
                    treeitem.TreeItem(obj=JsonItem(key=k, value=v, typ=type(v)))
                    for k, v in item.obj.value.items()
                ]

            case Iterable() if not isinstance(item.obj.value, str):
                return [
                    treeitem.TreeItem(obj=JsonItem(key=k, value=v, typ=type(v)))
                    for k, v in enumerate(item.obj.value)
                ]
            case _:
                return [
                    treeitem.TreeItem(
                        obj=JsonItem(
                            key="key",
                            value=repr(item.obj.value),
                            typ=type(item.obj.value),
                        )
                    )
                ]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    dist = [dict(a=2, b={"a": 4, "b": [1, 2, 3], "jkjkjk": "tekjk"}), 6, "jkjk"]
    model = JsonModel(dist)
    table = widgets.TreeView()
    table.setRootIsDecorated(True)
    # table.setSortingEnabled(True)
    table.set_model(model)
    table.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Callable
import html
import os
from typing import Literal

from prettyqt import constants, core, gui, iconprovider
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import bidict, datatypes, get_repr, prettyprinter


ACTION_EVENT = bidict(
    trigger=QtGui.QAction.ActionEvent.Trigger,
    hover=QtGui.QAction.ActionEvent.Hover,
)

ActionEventStr = Literal["trigger", "hover"]


PRIORITIES = bidict(
    low=QtGui.QAction.Priority.LowPriority,
    normal=QtGui.QAction.Priority.NormalPriority,
    high=QtGui.QAction.Priority.HighPriority,
)

PriorityStr = Literal["low", "normal", "high"]

ROLES = bidict(
    none=QtGui.QAction.MenuRole.NoRole,
    text_heuristic=QtGui.QAction.MenuRole.TextHeuristicRole,
    application_specific=QtGui.QAction.MenuRole.ApplicationSpecificRole,
    about_qt=QtGui.QAction.MenuRole.AboutQtRole,
    about=QtGui.QAction.MenuRole.AboutRole,
    preferences=QtGui.QAction.MenuRole.PreferencesRole,
    quit=QtGui.QAction.MenuRole.QuitRole,
)

RoleStr = Literal[
    "none",
    "text_heuristic",
    "application_specific",
    "about_qt",
    "about",
    "preferences",
    "quit",
]


class ActionMixin(core.ObjectMixin):
    def __init__(
        self,
        *args,
        text: str | None = None,
        icon: datatypes.IconType = None,
        callback: Callable | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._menu = None
        if callback is not None:
            self.triggered.connect(callback)
        self._usage_count = 0
        if text:
            self.setText(text)
        if icon:
            self.set_icon(icon)
        self.triggered.connect(self._increase_usage_counter)

    def __repr__(self) -> str:
        return get_repr(self, self.text())

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "priority": PRIORITIES,
            "shortcutContext": constants.SHORTCUT_CONTEXT,
            "menuRole": ROLES,
        }
        return maps

    def _increase_usage_counter(self):
        self._usage_count += 1

    def get_usage_count(self) -> int:
        return self._usage_count

    def get_type(self) -> Literal["menu", "separator", "widget", "regular"]:
        if self.menu() is not None:
            return "menu"
        elif self.isSeparator():
            return "separator"
        elif hasattr(self, "defaultWidget"):
            return "widget"
        else:
            return "regular"

    def set_disabled(self):
        self.setEnabled(False)

    def set_enabled(self, enabled: bool = True):
        self.setEnabled(enabled)

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        size: datatypes.SizeType | None = None,
        rich_text: bool = False,
    ):
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                if isinstance(size, QtCore.QSize):
                    size = (size.width(), size.height())
                tooltip = f'<img src={path!r} width="{size[0]}" height="{size[1]}">'
        tooltip = tooltip.replace("\n", "<br/>")
        if rich_text:
            tooltip = f"<html>{html.escape(tooltip)}</html>"
        super().setToolTip(tooltip)

    setToolTip = set_tooltip

    def set_icon(self, icon: datatypes.IconType):
        """Set the icon for the action.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        super().setIcon(icon)

    setIcon = set_icon

    def set_shortcut(self, shortcut: None | QtGui.QKeySequence | str):
        if shortcut is None:
            shortcut = ""
        if isinstance(shortcut, str):
            shortcut = gui.KeySequence(
                shortcut, gui.KeySequence.SequenceFormat.PortableText
            )
        super().setShortcut(shortcut)

    setShortcut = set_shortcut

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def get_shortcut(self) -> gui.KeySequence | None:
        shortcut = self.shortcut()
        return (
            gui.KeySequence(
                shortcut.toString(), gui.KeySequence.SequenceFormat.PortableText
            )
            if shortcut
            else None
        )

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def set_menu(self, menu):
        try:
            self.setMenu(menu)
        except AttributeError:
            self.triggered.connect(menu.exec)
            self._menu = menu

    def menu(self):
        return self._menu

    def set_priority(self, priority: PriorityStr | QtGui.QAction.Priority):
        """Set priority of the action.

        Args:
            priority: priority for the action

        Raises:
            InvalidParamError: priority does not exist
        """
        if isinstance(priority, str):
            priority = PRIORITIES[priority]
        super().setPriority(priority)

    setPriority = set_priority

    def get_priority(self) -> PriorityStr:
        """Return current priority.

        Returns:
            priority
        """
        return PRIORITIES.inverse[self.priority()]

    def set_shortcut_context(
        self, context: constants.ShortcutContextStr | QtCore.Qt.ShortcutContext
    ):
        """Set shortcut context.

        Args:
            context: shortcut context

        Raises:
            InvalidParamError: shortcut context does not exist
        """
        if isinstance(context, str):
            context = constants.SHORTCUT_CONTEXT[context]
        super().setShortcutContext(context)

    setShortcutContext = set_shortcut_context

    def get_shortcut_context(self) -> constants.ShortcutContextStr:
        """Return shortcut context.

        Returns:
            shortcut context
        """
        return constants.SHORTCUT_CONTEXT.inverse[super().shortcutContext()]

    def set_menu_role(self, role: RoleStr):
        """Set menu role.

        Args:
            role: menu role

        Raises:
            InvalidParamError: menu role does not exist
        """
        if isinstance(role, str):
            role = ROLES[role]
        super().setMenuRole(role)

    setMenuRole = set_menu_role

    def get_menu_role(self) -> RoleStr:
        """Return menu role.

        Returns:
            menu role
        """
        return ROLES.inverse[super().menuRole()]

    def show_shortcut_in_contextmenu(self, state: bool = True):
        self.setShortcutVisibleInContextMenu(state)

    usage_count = core.Property(int, get_usage_count)


class Action(ActionMixin, prettyprinter.PrettyPrinter, QtGui.QAction):
    pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    action = Action(text="This is a test", shortcut=None)



"""bluetooth module.

contains QtBluetooth-based classes
"""

from .bluetoothdeviceinfo import BluetoothDeviceInfo
from .bluetoothuuid import BluetoothUuid
from .bluetoothaddress import BluetoothAddress
from .bluetoothhostinfo import BluetoothHostInfo
from .bluetoothlocaldevice import BluetoothLocalDevice
from .bluetoothserviceinfo import BluetoothServiceInfo
from .bluetoothsocket import BluetoothSocket
from .bluetoothserver import BluetoothServer
from .bluetoothdevicediscoveryagent import BluetoothDeviceDiscoveryAgent
from .bluetoothservicediscoveryagent import BluetoothServiceDiscoveryAgent


__all__ = [
    "BluetoothHostInfo",
    "BluetoothLocalDevice",
    "BluetoothDeviceInfo",
    "BluetoothUuid",
    "BluetoothSocket",
    "BluetoothServer",
    "BluetoothAddress",
    "BluetoothServiceInfo",
    "BluetoothDeviceDiscoveryAgent",
    "BluetoothServiceDiscoveryAgent",
]



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtBluetooth
from prettyqt.utils import bidict


AttributeId = QtBluetooth.QBluetoothServiceInfo.AttributeId

ATTRIBUTE_IDS = bidict(
    service_record_handle=AttributeId.ServiceRecordHandle,
    service_class_ids=AttributeId.ServiceClassIds,
    service_record_state=AttributeId.ServiceRecordState,
    service_id=AttributeId.ServiceId,
    protocol_descriptor_list=AttributeId.ProtocolDescriptorList,
    browse_group_list=AttributeId.BrowseGroupList,
    language_base_attribute_id_list=AttributeId.LanguageBaseAttributeIdList,
    service_info_time_to_live=AttributeId.ServiceInfoTimeToLive,
    service_availablity=AttributeId.ServiceAvailability,
    bluetooth_profile_descriptor_list=AttributeId.BluetoothProfileDescriptorList,
    documentation_url=AttributeId.DocumentationUrl,
    client_executable_url=AttributeId.ClientExecutableUrl,
    icon_url=AttributeId.IconUrl,
    additional_protocol_descriptor_list=AttributeId.AdditionalProtocolDescriptorList,
    primary_language_base=AttributeId.PrimaryLanguageBase,
    # service_name=AttributeId.ServiceName,
    service_description=AttributeId.ServiceDescription,
    service_provider=AttributeId.ServiceProvider,
)

Protocol = QtBluetooth.QBluetoothServiceInfo.Protocol

PROTOCOL = bidict(
    unknown=Protocol.UnknownProtocol,
    l2_cap=Protocol.L2capProtocol,
    rfcomm=Protocol.RfcommProtocol,
)

ProtocolStr = Literal["unknown", "l2_cap", "rfcomm"]


class BluetoothServiceInfo(QtBluetooth.QBluetoothServiceInfo):
    def __getitem__(self, value: str):
        attr = ATTRIBUTE_IDS.inverse[value]
        return self.attribute(attr)

    def __delitem__(self, value: str):
        attr = ATTRIBUTE_IDS.inverse[value]
        return self.removeAttribute(attr)

    def __setitem__(self, index: str, value):
        attr = ATTRIBUTE_IDS.inverse[index]
        return self.setAttribute(attr, value)

    def __contains__(self, value: int) -> bool:
        attr = ATTRIBUTE_IDS.inverse[value]
        return self.contains(attr)

    def __iter__(self):
        return iter(self.attributes())


if __name__ == "__main__":
    address = BluetoothServiceInfo()



from __future__ import annotations

from prettyqt import constants, core, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError


class CheckBox(widgets.AbstractButtonMixin, QtWidgets.QCheckBox):
    value_changed = core.Signal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stateChanged.connect(self.value_changed)

    def set_checkstate(self, state: constants.CheckStateStr):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use

        Raises:
            InvalidParamError: invalid checkstate
        """
        if state not in constants.CHECK_STATE:
            raise InvalidParamError(state, constants.CHECK_STATE)
        self.setCheckState(constants.CHECK_STATE[state])

    def get_checkstate(self) -> constants.CheckStateStr:
        """Return checkstate.

        Returns:
            checkstate
        """
        return constants.CHECK_STATE.inverse[self.checkState()]


if __name__ == "__main__":
    app = widgets.app()
    widget = CheckBox("test")
    widget.show()
    app.main_loop()



from __future__ import annotations

import enum
import logging

from prettyqt import core, custom_widgets, eventfilters, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import bidict, datatypes, helpers


logger = logging.getLogger(__name__)


class WidgetEditor(widgets.ScrollArea):
    """Widget which automatically generates list of property editors for a given widget.

    We need to be careful here because we cannot rely on having all our mixin methods
    if we want this to work for all widgets.
    Would be worth investigating if we can "monkey-patch" the instances
    with self._widget.__class__ = type(cls.__name__,(OldClass, WidgetMixin),{})
    or self._widget.__bases__=(OldClass, WidgetMixin,)
    """

    value_changed = core.Signal(object)

    def __init__(self, widget: QtWidgets.QWidget, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._widget = widget
        self._initial_prop_values = {}
        self.event_catcher = eventfilters.EventCatcher(
            exclude=QtCore.QEvent.Type.Paint, parent=self._widget
        )
        self.event_catcher.caught.connect(self._update_editors)
        self._widget.installEventFilter(self.event_catcher)
        container = widgets.Widget()
        container.set_layout("form")
        self.setWidget(container)
        self.set_minimum_size(800, 1000)
        self._editors = bidict()
        # self.add_widget(container)
        self._metaobj = core.MetaObject(self._widget.metaObject())
        for i, prop in enumerate(self._metaobj.get_properties()):
            value = prop.read(self._widget)
            typ = prop.get_meta_type().get_type()
            name = prop.name()
            logger.info(f"setting {name} editor to {value}")
            if typ == bool:
                widget = widgets.CheckBox()
            elif typ == int:
                widget = widgets.SpinBox(maximum=999999)
            elif typ == float:
                widget = widgets.DoubleSpinBox(maximum=999999.0)
            elif typ == str:
                widget = widgets.LineEdit()
            elif typ == QtCore.QPoint:
                widget = custom_widgets.PointEdit()
            elif typ == QtCore.QRect:
                widget = custom_widgets.RectEdit()
            elif typ == QtGui.QRegion:
                widget = custom_widgets.RegionEdit()
            elif typ == QtCore.QSize:
                widget = custom_widgets.SizeEdit()
            elif typ == QtWidgets.QSizePolicy:
                widget = custom_widgets.SizePolicyEdit()
            elif typ == QtGui.QPalette:
                widget = custom_widgets.PaletteEdit()
            elif typ == QtGui.QIcon:
                widget = custom_widgets.IconEdit()
            elif typ == QtGui.QCursor:
                widget = custom_widgets.CursorEdit()
            elif typ == QtCore.QLocale:
                widget = custom_widgets.LocaleEdit()
            elif typ == QtGui.QFont:
                widget = widgets.FontComboBox()
            elif typ == QtGui.QColor:
                widget = custom_widgets.ColorComboBox()
            elif typ == enum.Flag:
                widget = custom_widgets.EnumFlagWidget()
            elif typ == enum.Enum and prop.isFlagType():
                widget = custom_widgets.EnumFlagWidget()
            elif typ == enum.Enum:
                widget = custom_widgets.EnumComboBox()
            else:
                raise ValueError(typ)
            widget.set_value(value)
            widget.value_changed.connect(self._on_value_change)
            widget.setEnabled(prop.isWritable())
            label = helpers.to_snake(prop.name()).replace("_", " ")
            container.box[i, "left"] = widgets.Label(label)
            container.box[i, "right"] = widget
            self._initial_prop_values[name] = value
            self._editors[name] = widget
            if prop.hasNotifySignal():
                notify_signal = prop.get_notify_signal()
                signal_name = notify_signal.get_name()
                signal = self._widget.__getattribute__(signal_name)
                signal.connect(self._update_editors)
            self.ensureWidgetVisible(container)
        self.setWidgetResizable(True)

    def _on_value_change(self):
        editor = self.sender()
        prop_name = self._editors.inverse[editor]
        prop = self._metaobj.get_property(prop_name)
        value = editor.get_value()
        value = datatypes.make_qtype(value)
        logger.info(f"setting {prop_name} to {value}")
        prop.write(self._widget, value)
        # brute force
        self._widget.updateGeometry()
        self._widget.repaint()
        if (parent := self._widget.parentWidget()) is not None:
            parent.updateGeometry()
            parent.repaint()

    def _update_editors(self):
        for i in range(self._widget.metaObject().propertyCount()):
            prop = self._widget.metaObject().property(i)
            prop_name = prop.name()
            editor = self._editors[prop_name]
            value = prop.read(self._widget)
            editor.set_value(value)


if __name__ == "__main__":
    app = widgets.app()
    container = widgets.Widget()
    container.set_layout("horizontal")
    w = widgets.PlainTextEdit(parent=container)
    w2 = widgets.PlainTextEdit(parent=container)
    container.box.add(w)
    container.box.add(w2)
    container.show()
    editor = WidgetEditor(w)
    editor.show()
    with app.debug_mode():
        app.main_loop()
    print(w.get_properties())



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import bidict


SELECTION_FLAG = bidict(
    none=QtCore.QItemSelectionModel.SelectionFlag.NoUpdate,
    clear=QtCore.QItemSelectionModel.SelectionFlag.Clear,
    select=QtCore.QItemSelectionModel.SelectionFlag.Select,
    deselect=QtCore.QItemSelectionModel.SelectionFlag.Deselect,
    toggle=QtCore.QItemSelectionModel.SelectionFlag.Toggle,
    current=QtCore.QItemSelectionModel.SelectionFlag.Current,
    rows=QtCore.QItemSelectionModel.SelectionFlag.Rows,
    columns=QtCore.QItemSelectionModel.SelectionFlag.Columns,
    select_current=QtCore.QItemSelectionModel.SelectionFlag.SelectCurrent,
    toggle_current=QtCore.QItemSelectionModel.SelectionFlag.ToggleCurrent,
    clear_and_select=QtCore.QItemSelectionModel.SelectionFlag.ClearAndSelect,
)

SelectionFlagStr = Literal[
    "none",
    "clear",
    "select",
    "deselect",
    "toggle",
    "current",
    "rows",
    "columns",
    "select_current",
    "toggle_current",
    "clear_and_select",
]


class ItemSelectionModel(core.ObjectMixin, QtCore.QItemSelectionModel):
    def set_current_index(self, index, flag: SelectionFlagStr):
        self.setCurrentIndex(index, SELECTION_FLAG[flag])



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


class ScrollArea(widgets.AbstractScrollAreaMixin, QtWidgets.QScrollArea):
    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def __add__(self, other: QtWidgets.QWidget | QtWidgets.QLayout | list):
        self.add(other)
        return self

    def get_children(self) -> list[QtWidgets.QWidget]:
        return self.widget().layout().get_children()

    def set_widget(self, widget: QtWidgets.QWidget):
        self.setWidget(widget)

    def add_widget(self, *args, **kwargs):
        self.widget().layout().addWidget(*args, **kwargs)

    def add_layout(self, *args, **kwargs):
        self.widget().layout().addLayout(*args, **kwargs)

    def add(
        self,
        item: QtWidgets.QWidget | QtWidgets.QLayout | list,
        stretch: float | None = None,
    ):
        match item:
            case QtWidgets.QWidget():
                self.add_widget(item)
                if stretch:
                    self.widget().layout().setStretchFactor(self.count() - 1, stretch)
            case QtWidgets.QLayout():
                widget = widgets.Widget(self)
                widget.set_layout(item)
                self.add_widget(widget)
                if stretch:
                    self.setStretchFactor(self.count() - 1, stretch)
            case list():
                for i in item:
                    self.add(i, stretch)
        return item


if __name__ == "__main__":
    app = widgets.app()
    area = ScrollArea()
    for i in range(50):
        w = widgets.RadioButton(str(i))
        area.setWidgetResizable(True)
        area.add_widget(w)
        area.ensureWidgetVisible(w)
    area.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtCore, QtGui, QtWidgets


class WidgetPropertiesModel(core.AbstractTableModel):
    HEADER = [
        "Property name",
        "Value",
        "Type",
        "User property",
        "Readable",
        "Writable",
        "Resettable",
    ]

    def __init__(self, widget: QtWidgets.QWidget, **kwargs):
        super().__init__(**kwargs)
        self._widget = widget
        self._metaobj = core.MetaObject(self._widget.metaObject())

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: QtCore.Qt.Orientation,
        role: QtCore.Qt.ItemDataRole,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self.HEADER[section]
            case constants.VERTICAL, constants.DISPLAY_ROLE:
                prop = self._metaobj.get_property(section)
                return prop.propertyIndex()

    def data(self, index, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        prop = self._metaobj.get_property(index.row())
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return prop.name()
            case constants.FONT_ROLE, 0:
                font = QtGui.QFont()
                font.setBold(True)
                return font
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE, 1:
                return prop.read(self._widget)
            case constants.DISPLAY_ROLE, 2:
                return prop.get_meta_type().get_name()
            case constants.FONT_ROLE, 2:
                font = QtGui.QFont()
                font.setItalic(True)
                return font
            case constants.CHECKSTATE_ROLE, 3:
                return prop.isUser()
            case constants.CHECKSTATE_ROLE, 4:
                return prop.isReadable()
            case constants.CHECKSTATE_ROLE, 5:
                return prop.isWritable()
            case constants.CHECKSTATE_ROLE, 6:
                return prop.isResettable()
            case constants.USER_ROLE, _:
                return prop.read(self._widget)

    def setData(self, index, value, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        prop = self._metaobj.get_property(index.row())
        match role, index.column():
            case constants.USER_ROLE, _:
                prop.write(self._widget, value)
                self.update_row(index.row())
                return True
        return False

    def rowCount(self, parent=None):
        """Override for AbstractitemModel base method."""
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        return 0 if parent.isValid() else self._metaobj.propertyCount()

    def flags(self, index):
        prop = self._metaobj.get_property(index.row())
        if index.column() == 1 and prop.isWritable():
            return (
                super().flags(index)
                | constants.IS_EDITABLE
                | constants.IS_ENABLED
                | constants.IS_SELECTABLE
            )
        return super().flags(index)


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt.custom_delegates import variantdelegate

    app = widgets.app()
    view = widgets.TableView()
    view.set_icon("mdi.folder")
    model = WidgetPropertiesModel(view)
    model.dataChanged.connect(view.repaint)
    delegate = variantdelegate.VariantDelegate(parent=view)
    view.set_model(model)
    view.set_selection_behavior("rows")
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_delegate(delegate, column=1)
    view.show()
    view.resize(500, 300)
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

import datetime
import enum
import logging
import pathlib

import regex as re

from prettyqt import constants, custom_delegates, custom_widgets, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets


logger = logging.getLogger(__name__)


class VariantDelegate(widgets.StyledItemDelegate):
    def __init__(
        self,
        data_role: QtCore.Qt.ItemDataRole = constants.USER_ROLE,
        parent: QtWidgets.QAbstractItemView | None = None,
    ):
        super().__init__(parent)
        self.data_role = data_role

    def paint(self, painter, option, index):
        value = self._data_for_index(index, self.data_role)
        # if not self.is_supported_type(value):
        #     option = widgets.StyleOptionViewItem(option)
        #     option.state &= ~QtWidgets.QStyle.StateFlag.State_Enabled
        if isinstance(value, QtGui.QIcon):
            icon_delegate = custom_delegates.IconDelegate()
            icon_delegate.paint(painter, option, index)
            return
        elif isinstance(value, enum.Enum):  # PySide6 needs this when using Views
            option.text = value.name
            option.widget.style().drawControl(
                QtWidgets.QStyle.ControlElement.CE_ItemViewItem, option, painter
            )
        else:
            super().paint(painter, option, index)

    def createEditor(self, parent, option, index):
        original_value = self._data_for_index(index, self.data_role)
        logger.info(f"creating editor for {original_value!r}...")
        match original_value:
            case bool():
                widget = widgets.CheckBox()
            case enum.Flag():
                widget = custom_widgets.EnumFlagWidget()
                widget._set_enum_class(type(original_value))
            case enum.Enum():
                widget = custom_widgets.EnumComboBox()
                widget._set_enum_class(type(original_value))
            case int():
                widget = widgets.SpinBox()
            case float():
                widget = widgets.DoubleSpinBox()
            case pathlib.Path():
                widget = custom_widgets.FileChooserButton()
            case str():
                widget = widgets.LineEdit()
                widget.setFrame(False)
            case QtCore.QRegularExpression() | re.Pattern():
                widget = custom_widgets.RegexInput(show_error=False)
            case QtCore.QTime():
                widget = widgets.TimeEdit()
            case QtCore.QDate():
                widget = widgets.DateEdit()
            case QtCore.QDateTime():
                widget = widgets.DateTimeEdit()
            case QtCore.QPoint():
                widget = custom_widgets.PointEdit()
            case QtCore.QSize():
                widget = custom_widgets.SizeEdit()
            case QtCore.QRect():
                widget = custom_widgets.RectEdit()
            case QtGui.QKeySequence():
                widget = widgets.KeySequenceEdit()
            case QtGui.QRegion():
                widget = custom_widgets.RegionEdit()
            case QtGui.QFont():
                widget = widgets.FontComboBox()
            case QtGui.QColor():
                widget = custom_widgets.ColorComboBox()
            case QtWidgets.QSizePolicy():
                widget = custom_widgets.SizePolicyEdit()
            # case QtCore.QRectF():  # todo
            #     widget = custom_widgets.RectEdit(parent=parent)
            case _:
                raise ValueError(original_value)
        widget.setParent(parent)
        widget.setAutoFillBackground(True)
        return widget

    def setEditorData(self, editor, index):
        # if not editor:
        #     return
        value = self._data_for_index(index, self.data_role)
        logger.info(f"setting data for {editor!r} to {value!r}")
        editor.set_value(value)

    def setModelData(self, editor, model, index):
        if (value := editor.get_value()) is not None:
            logger.info(f"setting data for {model!r} to {value!r}")
            model.setData(index, value, self.data_role)
            model.setData(index, self.display_text(value), constants.DISPLAY_ROLE)
            self.commitData.emit(editor)
            # self.closeEditor.emit(editor, self.EndEditHint.NoHint)

    @staticmethod
    def is_supported_type(value):
        return isinstance(
            value,
            bool
            | enum.Flag
            | enum.Enum
            | int
            | float
            | str
            | re.Pattern
            | pathlib.Path
            | QtCore.QRegularExpression
            | QtCore.QDate
            | QtCore.QDateTime
            | QtCore.QTime
            | QtCore.QRect
            | QtCore.QPoint
            | QtCore.QSize
            | QtCore.QTime
            | QtGui.QPalette
            | QtGui.QIcon
            | QtGui.QColor
            | QtGui.QFont
            | QtGui.QKeySequence
            | QtWidgets.QSizePolicy,
        )

    @staticmethod
    def display_text(val):
        match val:
            case str():
                return val
            case bool():
                return "✓" if val else "☐"
            case enum.Flag():
                return val.name
            case enum.Enum():
                return val.name
            case int() | float() | QtCore.QByteArray():
                return str(val)
            case QtGui.QColor():
                return f"({val.red()},{val.green()},{val.blue()},{val.alpha()})"
            case QtGui.QFont():
                return val.family()
            case QtGui.QRegion():
                rect = val.boundingRect()
                return f"({rect.x()},{rect.y()},{rect.width()},{rect.height()})"
            case QtGui.QCursor():
                return constants.CURSOR_SHAPE.inverse[val.shape()]
            case QtGui.QKeySequence():
                return val.toString()
            case QtCore.QDate() | QtCore.QDateTime() | QtCore.QTime():
                return val.toString(QtCore.Qt.DateFormat.ISODate)
            case QtCore.QPoint():
                return f"({val.x()},{val.y()})"
            case QtCore.QRect():
                return f"({val.x()},{val.y()},{val.width()},{val.height()})"
            case QtCore.QSize():
                return f"({val.width()},{val.height()})"
            case QtCore.QLocale():
                return val.bcp47Name()
            case QtWidgets.QSizePolicy():
                return (
                    f"({widgets.sizepolicy.SIZE_POLICY.inverse[val.horizontalPolicy()]}, "
                    f"{widgets.sizepolicy.SIZE_POLICY.inverse[val.verticalPolicy()]}, "
                    f"{widgets.sizepolicy.CONTROL_TYPE.inverse[val.controlType()]})"
                )
            case list():
                return ",".join(map(repr, val))
            case re.Pattern():
                return val.pattern
            case datetime.date():
                return val.isoformat()
            case datetime.datetime():
                return val.isoformat(sep=" ")
            case QtCore.QRegularExpression():
                return val.pattern()
            case QtCore.QUrl():
                return val.toString()
            # case np.integer():
            #     return super().displayText(int(val), locale)
            # case np.floating():
            #     return super().displayText(float(val), locale)
            # case np.str_():
            #     return str(val)
            # case np.datetime64():
            #     return self.displayText(val.astype(datetime), locale)
            case None:
                return "<None>"
            case _:
                return f"<{val}>"

    def displayText(self, value, locale: QtCore.QLocale) -> str:
        return self.display_text(value)


if __name__ == "__main__":
    app = widgets.app()
    table_widget = widgets.TableWidget(15, 2)
    # table_widget.set_delegate(StarDelegate(), column=1)
    table_widget.setEditTriggers(
        table_widget.EditTrigger.DoubleClicked | table_widget.EditTrigger.SelectedClicked
    )
    table_widget.set_selection_behavior("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Rating"])
    types = dict(
        color=QtGui.QColor(30, 30, 30),
        time=QtCore.QTime(1, 1, 1),
        date=QtCore.QDate(1, 1, 1),
        datetime=QtCore.QDateTime(1, 1, 1, 1, 1, 1),
        font=QtGui.QFont(),
        str="fdsf",
        int=8,
        float=8.2,
        path=pathlib.Path(),
        url=QtCore.QUrl("http://www.google.de"),
        # regex=QtCore.QRegularExpression("[a-z]"),
        regex=re.compile("[a-z]"),
        enum=QtCore.Qt.ItemDataRole.UserRole,
        bool=True,
        keysequence=QtGui.QKeySequence("Ctrl+A"),
    )
    for i, (k, v) in enumerate(types.items()):
        item_1 = widgets.TableWidgetItem(k)
        item_2 = widgets.TableWidgetItem()
        item_2.setData(constants.DISPLAY_ROLE, v)
        item_2.setData(constants.USER_ROLE, v)
        table_widget[i, 0] = item_1
        table_widget[i, 1] = item_2

    delegate = VariantDelegate(parent=table_widget)
    table_widget.set_delegate(delegate, column=1)

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

from prettyqt import constants, gui
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError


class TextTableFormat(gui.textframeformat.TextFrameFormatMixin, QtGui.QTextTableFormat):
    def __bool__(self):
        return self.isValid()

    def set_alignment(self, alignment: constants.AlignmentStr):
        """Set the alignment of the format.

        Args:
            alignment: alignment for the format

        Raises:
            InvalidParamError: alignment does not exist
        """
        if alignment not in constants.ALIGNMENTS:
            raise InvalidParamError(alignment, constants.ALIGNMENTS)
        self.setAlignment(constants.ALIGNMENTS[alignment])

    def get_alignment(self) -> constants.AlignmentStr:
        """Return current alignment.

        Returns:
            alignment
        """
        return constants.ALIGNMENTS.inverse[self.alignment()]

    def get_column_width_constraints(self) -> list[gui.TextLength]:
        return [gui.TextLength(i) for i in self.columnWidthConstraints()]


if __name__ == "__main__":
    fmt = TextTableFormat()
    print(fmt)



from __future__ import annotations

from typing_extensions import Self

from prettyqt.qt import QtCore


class Time(QtCore.QTime):
    def __repr__(self):
        template = super().__repr__().split("(")[1]  # type: ignore
        return f"{type(self).__name__}({template}"

    def __str__(self):
        return self.toString()

    def __reduce__(self):
        return type(self), (self.hour(), self.minute(), self.second(), self.msec())

    def __format__(self, format_spec: str):
        return self.toString(format_spec)

    @property
    def _hour(self):
        return self.hour()

    @property
    def _minute(self):
        return self.minute()

    @property
    def _second(self):
        return self.second()

    @property
    def _msec(self):
        return self.msec()

    __match_args__ = ("_hour", "_minute", "_second", "_msec")

    def add_msecs(self, msecs: int) -> Self:
        return type(self)(self.addMSecs(msecs))

    def add_secs(self, secs: int) -> Self:
        return type(self)(self.addSecs(secs))

    @classmethod
    def get_current_time(cls) -> Self:
        return cls(cls.currentTime())


if __name__ == "__main__":
    time = Time(22, 1)
    print(time.add_secs(2))
    print(repr(time))



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


class HBoxLayout(widgets.boxlayout.BoxLayoutMixin, QtWidgets.QHBoxLayout):
    pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    layout = HBoxLayout()
    widget = widgets.Widget()
    widget2 = widgets.RadioButton("Test")
    layout.add(widget2)
    widget.set_layout(layout)
    widget.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Iterator

from prettyqt import constants, core, gui, widgets
from prettyqt.qt import QtWidgets


class FaderWidget(widgets.Widget):
    pixmap_opacity = 1.0

    def __init__(
        self,
        old_widget: QtWidgets.QWidget,
        new_widget: QtWidgets.QWidget,
        duration: int = 300,
    ):
        super().__init__(new_widget)

        pr = gui.Window().devicePixelRatio()
        self.old_pixmap = gui.Pixmap(new_widget.size() * pr)
        self.old_pixmap.setDevicePixelRatio(pr)
        old_widget.render(self.old_pixmap)

        self.timeline = core.TimeLine(duration=duration, finished=self.close)
        self.timeline.value_changed.connect(self.animate)
        self.timeline.start()

        self.resize(new_widget.size())
        self.show()

    def paintEvent(self, event):
        with gui.Painter(self) as painter:
            painter.setOpacity(self.pixmap_opacity)
            painter.drawPixmap(0, 0, self.old_pixmap)

    def animate(self, value: float):
        self.pixmap_opacity = 1.0 - value
        self.repaint()


class StackedWidget(widgets.FrameMixin, QtWidgets.QStackedWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.direction = "horizontal"
        self.speed = 300
        self.animation_type = "out_cubic"
        self.now = 0
        self.next = 0
        self.wrap = False
        self.pnow = core.Point(0, 0)
        self.active = False

    def __add__(self, other: QtWidgets.QWidget) -> StackedWidget:
        self.addWidget(other)
        return self

    def __getitem__(self, index: int) -> QtWidgets.QWidget:
        return self.widget(index)

    def __delitem__(self, item: int | QtWidgets.QWidget):
        if isinstance(item, int):
            item = self.widget(item)
        self.removeWidget(item)

    def __iter__(self) -> Iterator[QtWidgets.QWidget]:
        return iter(self.widget(i) for i in range(self.count()))

    def __len__(self):
        # needed for PySide2
        return self.count()

    def __contains__(self, item: QtWidgets.QWidget):
        return self.indexOf(item) >= 0

    def set_direction(self, direction: constants.OrientationStr):
        self.direction = direction

    def set_speed(self, speed: int):
        self.speed = speed

    def set_animation(self, animation_type: core.easingcurve.TypeStr):
        self.animation_type = animation_type

    def set_wrap(self, wrap: bool):
        self.wrap = wrap

    @core.Slot()
    def slide_in_previous(self):
        now = self.currentIndex()
        if self.wrap or now > 0:
            self.slide_in(now - 1)

    @core.Slot()
    def slide_in_next(self):
        now = self.currentIndex()
        if self.wrap or now < (self.count() - 1):
            self.slide_in(now + 1)

    def slide_in(self, idx: int | QtWidgets.QWidget):
        if self.active:
            return
        if isinstance(idx, int):
            if idx > (self.count() - 1):
                idx = idx % self.count()
            elif idx < 0:
                idx = (idx + self.count()) % self.count()
            idx = self.widget(idx)

        self.active = True

        _now = self.currentIndex()
        _next = self.indexOf(idx)

        if _now == _next:
            self.active = False
            return

        offsetx, offsety = self.frameRect().width(), self.frameRect().height()
        self.widget(_next).setGeometry(self.frameRect())

        if self.direction == "horizontal":
            if _now < _next:
                offsetx, offsety = -offsetx, 0
            else:
                offsety = 0

        elif _now < _next:
            offsetx, offsety = 0, -offsety
        else:
            offsetx = 0
        pnext = self.widget(_next).pos()
        pnow = self.widget(_now).pos()
        self.pnow = pnow

        offset = core.Point(offsetx, offsety)
        self.widget(_next).move(pnext - offset)
        self.widget(_next).show()
        self.widget(_next).raise_()

        anim_group = core.ParallelAnimationGroup(self, finished=self._on_animation_done)

        for index, start, end in zip(
            (_now, _next), (pnow, pnext - offset), (pnow + offset, pnext)
        ):
            animation = core.PropertyAnimation(
                self.widget(index),
                b"pos",
                duration=self.speed,
                easing_curve=self.animation_type,
                start_value=start,
                end_value=end,
            )
            anim_group.addAnimation(animation)

        self.next = _next
        self.now = _now
        self.active = True
        anim_group.start_animation(policy="delete")

    @core.Slot()
    def _on_animation_done(self):
        self.setCurrentIndex(self.next)
        self.widget(self.now).hide()
        self.widget(self.now).move(self.pnow)
        self.active = False

    def fade_in(self, widget: int | QtWidgets.QWidget):
        widget = self.widget(widget) if isinstance(widget, int) else widget
        self.fader_widget = FaderWidget(self.currentWidget(), widget, self.speed)
        self.setCurrentIndex(widget)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    stackedwidget = StackedWidget()
    widget2 = widgets.RadioButton("Test")
    widget3 = widgets.RadioButton("Test 2")
    stackedwidget += widget2
    stackedwidget += widget3
    stackedwidget.show()
    app.sleep(3)
    stackedwidget.slide_in_next()
    app.main_loop()



from __future__ import annotations

from prettyqt import core, iconprovider, widgets
from prettyqt.qt import QtWidgets


class PopupInfo(widgets.Dialog):
    """Dialog overlay to show some info to user."""

    def __init__(self, parent: QtWidgets.QWidget | None = None, text: str | None = None):
        super().__init__(parent=parent)
        self.timer = core.Timer(single_shot=True, timeout=self.close)
        self.label = widgets.Label(alignment="center")
        self.iconlabel = widgets.Label()
        self.set_icon("mdi.information")
        self.set_flags(stay_on_top=True, frameless=True, tooltip=True)
        self.hide()
        layout = widgets.HBoxLayout(margin=20, size_constraint="minimum")
        self.set_layout(layout)
        self.set_background_color("black")
        self.label.set_color("white")
        layout.add(self.iconlabel)
        layout.add(self.label)

    def set_icon(self, icon: str):
        pixmap = iconprovider.get_icon(icon, color="white").pixmap(32)
        self.iconlabel.setPixmap(pixmap)

    def show(self):
        self.adjustSize()
        self.position_on("screen", y_offset=-200)
        super().show()
        self.timer.start(2500)

    def show_popup(self, text: str):
        self.label.setText(text)
        self.show()


if __name__ == "__main__":
    app = widgets.app()
    widget = PopupInfo()
    widget.show_popup("tfsdfdsfdsfsdfsdest")
    app.main_loop()



from __future__ import annotations

from collections.abc import Iterable
import re
from typing import Literal

from prettyqt import constants, core
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError


class SortFilterProxyModel(core.AbstractProxyModelMixin, QtCore.QSortFilterProxyModel):
    invalidated = core.Signal()

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self._filter_column = 0

    # def setFilterKeyColumn(self, column: int | list[int] | None):
    #     if column is None:
    #         column = -1
    #     self._filter_column = column

    # def filterKeyColumn(self) -> int | list[int] | None:
    #     return self._filter_column

    # def filterAcceptsRow(self, source_row: int, source_index: core.ModelIndex):
    #     column = self.filterKeyColumn()
    #     col_count = self.sourceModel().columnCount()
    #     indexes = [
    #         self.sourceModel().index(source_row, i, source_index) for i in col_count
    #     ]
    #     labels = [self.sourceModel().data(idx) for idx in indexes]

    #     if isinstance(column, int) and source_index.column() == column:
    #         return super().filterAcceptsRow(source_row, source_index)

    def invalidate(self):
        super().invalidate()
        self.invalidated.emit()

    def lessThan(self, left, right):
        role = super().sortRole()
        left_data = left.data(role)
        right_data = right.data(role)
        if left_data is not None and right_data is not None:
            return left_data < right_data
        return True

    def sort(
        self,
        column: int | None,
        ascending: bool | QtCore.Qt.SortOrder = constants.ASCENDING,
    ):
        if isinstance(ascending, bool):
            ascending = constants.ASCENDING if ascending else constants.DESCENDING
        if column is None:
            column = -1
        super().sort(column, ascending)

    def setFilterString(self, search_str: str):
        pat = ".*?".join(map(re.escape, search_str))
        pat = f"(?=({pat}))"
        super().setFilterRegularExpression(pat)

    def setFixedFilterList(self, filter_list: Iterable[str]):
        pat = "|".join(filter_list)
        super().setFilterRegularExpression(pat)

    def set_filter_case_sensitivity(self, sensitivity: constants.CaseSensitivityStr):
        """Set the filter case sensitivity.

        Args:
            sensitivity: filter case sensitivity

        Raises:
            InvalidParamError: filter case sensitivity does not exist
        """
        if sensitivity not in constants.CASE_SENSITIVITY:
            raise InvalidParamError(sensitivity, constants.CASE_SENSITIVITY)
        super().setFilterCaseSensitivity(constants.CASE_SENSITIVITY[sensitivity])

    def get_filter_case_sensitivity(self) -> constants.CaseSensitivityStr:
        """Return current filter case sensitivity.

        Returns:
            filter case sensitivity
        """
        return constants.CASE_SENSITIVITY.inverse[super().filterCaseSensitivity()]

    def set_filter_case_sensitive(self, state: bool):
        if state:
            sensitivity = QtCore.Qt.CaseSensitivity.CaseSensitive
        else:
            sensitivity = QtCore.Qt.CaseSensitivity.CaseInsensitive
        super().setFilterCaseSensitivity(sensitivity)

    def is_filter_case_sensitive(self) -> bool:
        return super().filterCaseSensitivity() == QtCore.Qt.CaseSensitivity.CaseSensitive

    def set_sort_case_sensitive(self, state: bool):
        if state:
            sensitivity = QtCore.Qt.CaseSensitivity.CaseSensitive
        else:
            sensitivity = QtCore.Qt.CaseSensitivity.CaseInsensitive
        super().setSortCaseSensitivity(sensitivity)

    def is_sort_case_sensitive(self) -> bool:
        return super().sortCaseSensitivity() == QtCore.Qt.CaseSensitivity.CaseSensitive

    def get_filter_regular_expression(self) -> core.RegularExpression:
        return core.RegularExpression(self.filterRegularExpression())

    def set_sort_role(self, role: constants.ItemDataRoleStr | int):
        role = constants.ITEM_DATA_ROLE[role] if isinstance(role, str) else role
        super().setSortRole(role)

    def set_filter_role(self, role: constants.ItemDataRoleStr | int):
        role = constants.ITEM_DATA_ROLE[role] if isinstance(role, str) else role
        super().setFilterRole(role)

    def get_sort_order(self) -> Literal["ascending", "descending"]:
        return "ascending" if super().sortOrder() == constants.ASCENDING else "descending"

    def set_filter_key_column(self, column: int | None):
        if column is None:
            column = -1
        super().setFilterKeyColumn(column)


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt.custom_models import JsonModel

    app = widgets.app()
    dist = [
        dict(
            a=2,
            b={
                "a": 4,
                "b": [1, 2, 3],
                "jkjkjk": "tekjk",
                "sggg": "tekjk",
                "fdfdf": "tekjk",
                "xxxx": "axxxb",
            },
        ),
        6,
        "jkjk",
    ]
    source_model = JsonModel(dist)
    from prettyqt.custom_models.importlibdistributionmodel import (
        ImportlibDistributionModel,
    )

    source_model = ImportlibDistributionModel.from_package("prettyqt")
    model = SortFilterProxyModel()
    model.setSourceModel(source_model)
    lineedit = widgets.LineEdit()
    # completer = SubsequenceCompleter()
    # completer.setModel(source_model)
    # lineedit.set_completer(completer)
    widget = widgets.Widget()
    widget.set_layout("vertical")
    widget.box.add(lineedit)
    table = widgets.TableView()
    table.setSortingEnabled(True)
    widget.box.add(table)
    table.set_model(model)
    widget.show()
    app.main_loop()



"""Provides Qt init stuff."""

from typing import Literal

import warnings
import os
import sys


class PythonQtError(ImportError):
    pass


API_NAMES = {"pyqt6": "PyQt6", "pyside6": "PySide6"}


DEFAULT = "pyqt6"

# Detecting if a binding was specified by the user
binding_specified = "QT_API" in os.environ
API = os.environ.get("QT_API", DEFAULT).lower()

if API.lower() in {"pyqt5", "pyside2"}:
    API = DEFAULT
    binding_specified = False

initial_api = API
if API not in API_NAMES:
    raise ValueError(
        f"Specified QT_API={repr(API)} is not in valid options: " f"{API_NAMES}"
    )


PYQT6 = False
PYSIDE6 = False

API_VERSION = None
QT_VERSION = None


if "PyQt6" in sys.modules:
    API = "pyqt6"
elif "PySide6" in sys.modules:
    API = "pyside6"

if API == "pyqt6":
    try:
        from PyQt6.QtCore import PYQT_VERSION_STR as API_VERSION  # analysis:ignore
        from PyQt6.QtCore import QT_VERSION_STR as QT_VERSION  # analysis:ignore

        PYQT6 = True

    except ImportError:
        API = "pyside6"
    # else:
    #     os.environ["QT_API"] = API

if API == "pyside6":
    try:
        from PySide6 import __version__ as API_VERSION  # analysis:ignore
        from PySide6.QtCore import __version__ as QT_VERSION  # analysis:ignore

        PYSIDE6 = True

    except ImportError:
        raise QtBindingsNotFoundError from None
    # else:
    #     os.environ["QT_API"] = API


# If a correct API name is passed to QT_API and it could not be found,
# switches to another and informs through the warning
if API != initial_api and binding_specified:
    warnings.warn(
        f"Selected binding {initial_api!r} could not be found; "
        f"falling back to {API!r}",
        PythonQtWarning,
    )


# Set display name of the Qt API
API_NAME = API_NAMES[API]


def set_env_vars(qt_binding: Literal["PyQt6", "PySide6"]):
    ENV_VARS = ["QT_API", "USE_QT_API", "PYTEST_QT_API", "PYQTGRAPH_QT_LIB"]
    for var in ENV_VARS:
        os.environ[var] = qt_binding


set_env_vars(API_NAME)
